# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""General fermionic simulator backend tests."""

from time import sleep
import numpy as np

from qiskit_aer import AerJob
from qiskit import QuantumCircuit, QuantumRegister
from qiskit.providers import JobStatus
from qiskit.result import Result
from qiskit.test import QiskitTestCase

from qiskit_nature.operators.second_quantization import FermionicOp

from qiskit_cold_atom.exceptions import QiskitColdAtomError
from qiskit_cold_atom.fermions.fermion_simulator_backend import FermionSimulator
from qiskit_cold_atom.fermions.base_fermion_backend import BaseFermionBackend


class TestFermionSimulatorBackend(QiskitTestCase):
    """class to test the FermionSimulatorBackend class."""

    def setUp(self):
        super().setUp()
        self.backend = FermionSimulator()

    def test_initialization(self):
        """test the initialization of the backend"""
        target_config = {
            "backend_name": "fermion_simulator",
            "backend_version": "0.0.1",
            "n_qubits": 20,
            "basis_gates": None,
            "gates": [],
            "local": False,
            "simulator": True,
            "conditional": False,
            "open_pulse": False,
            "memory": True,
            "max_shots": 1e5,
            "coupling_map": None,
            "description": r"a base simulator for fermionic circuits. Instead of qubits, "
            r"each wire represents a single fermionic mode",
        }

        backend = FermionSimulator()
        self.assertIsInstance(backend, BaseFermionBackend)
        self.assertTrue(target_config.items() <= backend.configuration().to_dict().items())

    def test_run_method(self):
        """Test the run method of the backend simulator"""

        with self.subTest("test call"):
            circ = self.backend.initialize_circuit([[0, 1], [1, 0]])
            job = self.backend.run(circ)
            self.assertIsInstance(job, AerJob)
            self.assertIsInstance(job.job_id(), str)
            self.assertIsInstance(job.result(), Result)
            sleep(0.01)
            self.assertEqual(job.status(), JobStatus.DONE)

        circ1 = self.backend.initialize_circuit([[0, 1], [1, 0]])
        circ2 = self.backend.initialize_circuit([[1, 1], [1, 0]])

        with self.subTest("test call with multiple circuits"):
            job = self.backend.run([circ1, circ2])
            self.assertIsInstance(job, AerJob)

        with self.subTest("test shot number"):
            target_shots = 123
            job = self.backend.run([circ1, circ2], shots=target_shots)
            for exp in job.result().results:
                self.assertEqual(exp.shots, target_shots)

        with self.subTest("test seed of RNG"):
            target_seed = 123
            job = self.backend.run([circ1, circ2], seed=target_seed)
            for exp in job.result().results:
                self.assertEqual(exp.header.random_seed, target_seed)

        with self.subTest("test number of fermionic species"):
            # define a circuit that conserves the particle number per fermionic spin species
            test_circ = QuantumCircuit(4)
            test_circ.fload([0, 3])
            test_circ.fhop([np.pi / 4], [0, 1, 2, 3])

            statevector_1 = self.backend.run(test_circ).result().get_statevector()
            self.assertEqual(len(statevector_1), 6)
            # check whether specifying the number of species reduces the dimension of the simulation
            statevector_2 = self.backend.run(test_circ, num_species=2).result().get_statevector()
            self.assertEqual(len(statevector_2), 4)

    def test_execute(self):
        """test the ._execute() method internally called by .run()"""

        with self.subTest("test partial measurement"):
            circ_meas = QuantumCircuit(2, 2)
            circ_meas.fload(0)
            circ_meas.measure(0, 0)
            with self.assertWarns(UserWarning):
                self.backend.run(circ_meas)

        test_circ = QuantumCircuit(4)
        test_circ.fload([0, 3])
        test_circ.fhop([np.pi / 4], [0, 1, 2, 3])
        test_circ.fint(np.pi, [0, 1, 2, 3])
        test_circ.measure_all()

        result = self.backend.run(test_circ, num_species=2, seed=40, shots=5).result()

        with self.subTest("test simulation counts"):
            self.assertEqual(result.get_counts(), {"1010": 1, "0110": 3, "0101": 1})

        with self.subTest("test simulation memory"):
            self.assertEqual(result.get_memory(), ["0110", "0101", "1010", "0110", "0110"])

        with self.subTest("test simulation statevector"):
            self.assertTrue(
                np.allclose(result.get_statevector(), np.array([-0.5j, -0.5, 0.5, -0.5j]))
            )

        with self.subTest("test simulation unitary"):
            self.assertTrue(
                np.allclose(
                    result.get_unitary(),
                    np.array(
                        [
                            [-0.5, -0.5j, -0.5j, 0.5],
                            [0.5j, 0.5, -0.5, 0.5j],
                            [0.5j, -0.5, 0.5, 0.5j],
                            [0.5, -0.5j, -0.5j, -0.5],
                        ]
                    ),
                )
            )

        with self.subTest("test time taken"):
            self.assertTrue(result.to_dict()["time_taken"] < 0.1)

        with self.subTest("test result success"):
            self.assertTrue(result.to_dict()["success"])

    def test_initialize_circuit(self):
        """test of initialize_circuit inherited from the abstract base class BaseFermionBackend"""

        with self.subTest("Initialize circuit with single species of fermions"):
            actual_circ = self.backend.initialize_circuit([0, 1, 0, 1])
            target_circ = QuantumCircuit(QuantumRegister(4, "fer_mode"))
            target_circ.fload(1)
            target_circ.fload(3)
            self.assertEqual(actual_circ, target_circ)

        with self.subTest("Initialize circuit with multiple species of fermions"):
            actual_circ = self.backend.initialize_circuit([[0, 1], [0, 1]])
            target_circ = QuantumCircuit(QuantumRegister(2, "spin_0"), QuantumRegister(2, "spin_1"))
            target_circ.fload(1)
            target_circ.fload(3)
            self.assertEqual(actual_circ, target_circ)

        with self.subTest("check maximum size of circuit"):
            with self.assertRaises(QiskitColdAtomError):
                self.backend.initialize_circuit(np.ones(30, dtype=int).tolist())

    def test_measure_observable_expectation(self):
        """test of the measure_observable_expectation method inherited from the abstract base class
        BaseFermionBackend"""

        with self.subTest("test error for non-diagonal observables"):
            non_diag_observable = FermionicOp("+-NI")
            test_circ = self.backend.initialize_circuit([0, 1, 0, 1])
            with self.assertRaises(QiskitColdAtomError):
                self.backend.measure_observable_expectation(
                    test_circ, non_diag_observable, shots=10
                )

        with self.subTest("test match of dimensionality"):
            observable_too_small = FermionicOp("NI")
            test_circ = self.backend.initialize_circuit([0, 1, 0, 1])
            with self.assertRaises(QiskitColdAtomError):
                self.backend.measure_observable_expectation(
                    test_circ, observable_too_small, shots=10
                )

        with self.subTest("test single measurement circuit"):
            observable_1 = FermionicOp("INEI")
            observable_2 = FermionicOp("INII") + FermionicOp("IIEI")
            observable_3 = FermionicOp("NNII")

            eval_1 = self.backend.measure_observable_expectation(
                circuits=self.backend.initialize_circuit([0, 1, 0, 1]),
                observable=observable_1,
                shots=1,
            )
            eval_2 = self.backend.measure_observable_expectation(
                circuits=self.backend.initialize_circuit([0, 1, 0, 1]),
                observable=observable_2,
                shots=1,
            )
            eval_3 = self.backend.measure_observable_expectation(
                circuits=self.backend.initialize_circuit([0, 1, 0, 1]),
                observable=observable_3,
                shots=1,
            )

            self.assertEqual(eval_1, [1.0])
            self.assertEqual(eval_2, [2.0])
            self.assertEqual(eval_3, [0.0])

        with self.subTest("test multiple measurement circuits"):
            test_circ_1 = self.backend.initialize_circuit([0, 1, 0, 1])
            test_circ_2 = self.backend.initialize_circuit([1, 0, 0, 1])
            observable = FermionicOp("INEI")
            expval = self.backend.measure_observable_expectation(
                [test_circ_1, test_circ_2], observable, shots=1
            )
            self.assertEqual(expval, [1.0, 0.0])

    def test_parameterized_circuits(self):
        """Test that parameterized circuits work."""
        from qiskit.circuit import Parameter

        theta = Parameter("theta")

        test_circ = QuantumCircuit(4)
        test_circ.fload([0, 3])
        test_circ.fhop([theta], [0, 1, 2, 3])

        with self.subTest("test running with unbound parameters:"):
            with self.assertRaises(TypeError):
                self.assertTrue(isinstance(self.backend.run(test_circ).result(), Result))

        with self.subTest("test running with bound parameters"):
            bound_circ = test_circ.bind_parameters([0.2])
            self.assertTrue(isinstance(self.backend.run(bound_circ).result(), Result))



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""A base class for circuit solvers of cold atomic quantum circuits."""

from abc import ABC, abstractmethod
from typing import List, Optional, Dict, Any
import warnings
import numpy as np
from scipy.sparse import csc_matrix, identity, SparseEfficiencyWarning
from scipy.sparse.linalg import expm

from qiskit import QuantumCircuit
from qiskit.circuit import Gate
from qiskit_nature.operators.second_quantization import SecondQuantizedOp
from qiskit_cold_atom.exceptions import QiskitColdAtomError


class BaseCircuitSolver(ABC):
    """An abstract class for circuit solvers of different cold atom types.

    By subclassing BaseCircuitSolver one can create circuit solvers for different
    types of cold atomic setups such as spin, fermionic, and bosonic setups. All
    these subclasses will simulate cold atom quantum circuits by exponentiating
    matrices. Therefore, subclasses of BaseCircuitSolver are not intended to solve
    large circuits.
    """

    def __init__(
        self,
        shots: Optional[int] = None,
        seed: Optional[int] = None,
        max_dimension: int = 1e6,
        ignore_barriers: bool = True,
    ):
        """
        Args:
            shots: amount of shots for the measurement simulation;
                   if not None, measurements are performed, otherwise no measurements are done.
            seed: seed for the RNG for the measurement simulation
            max_dimension: The maximum Hilbert space dimension (limited to keep
                computation times reasonably short)
            ignore_barriers: If true, will ignore barrier instructions
        """
        self.shots = shots

        self._seed = seed
        if self._seed is not None:
            np.random.seed(self._seed)

        self._max_dimension = max_dimension
        self._ignore_barriers = ignore_barriers
        self._dim = None

    @property
    def seed(self) -> int:
        """The seed for the random number generator of the measurement simulation."""
        return self._seed

    @seed.setter
    def seed(self, value: int):
        """Set the seed for the random number generator. This will also update numpy's seed."""
        np.random.seed(value)
        self._seed = value

    @property
    def max_dimension(self) -> int:
        """The maximal Hilbert space dimension of the simulation."""
        return self._max_dimension

    @max_dimension.setter
    def max_dimension(self, value: int):
        self._max_dimension = value

    @property
    def ignore_barriers(self) -> bool:
        """Boolean flag that defines how barrier instructions in the circuit are handled."""
        return self._ignore_barriers

    @property
    def dim(self) -> int:
        """Return the dimension set by the last quantum circuit on which the solver was called."""
        return self._dim

    @ignore_barriers.setter
    def ignore_barriers(self, boolean):
        self._ignore_barriers = boolean

    def __call__(self, circuit: QuantumCircuit) -> Dict[str, Any]:
        """
        Performs the simulation of the circuit: Each operator is converted into a sparse matrix
        over the basis and is then exponentiated to get the unitary of the gate. All these
        unitaries are multiplied to give the total unitary of the circuit. Applying this to the
        initial state yields the final state of the circuit, from which we sample a number `shots`
        of shots (if specified).

        Args:
            circuit: A quantum circuit with gates described by second quantized generators

        Returns:
            output: dict{'unitary' : np.array((dimension, dimension)),
                 'statevector': np.array((dimension, 1)),
                 'counts': dict{string: int}}

        Raises:
            QiskitColdAtomError:
                - If one of the generating Hamiltonians is not hermitian which would
                lead to non-unitary time evolution.
                - If the dimension of the Hilbert space is larger than the max. dimension.
            NotImplementedError:
                - If ignore_barriers is False.
        """

        self.preprocess_circuit(circuit)

        if self._dim > self.max_dimension:
            raise QiskitColdAtomError(
                f"Hilbert space dimension of the simulation ({self._dim}) exceeds the "
                f"maximally supported value {self.max_dimension}."
            )

        # initialize the circuit unitary as an identity matrix
        circuit_unitary = identity(self._dim, dtype=complex)

        for op in self.to_operators(circuit):
            operator_mat = self.operator_to_mat(op)

            # check that the operators are hermitian before exponentiating
            if (operator_mat.H - operator_mat).count_nonzero() != 0:
                raise QiskitColdAtomError("generator of unitary gate is not hermitian!")
            # with the next release of qiskit nature this can be replaced with
            # if not operator.is_hermitian():
            #     raise QiskitColdAtomError("generator of unitary gate is not hermitian!")

            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=SparseEfficiencyWarning)
                gate_unitary = expm(-1j * operator_mat)

            circuit_unitary = gate_unitary @ circuit_unitary

        final_state = circuit_unitary @ self.get_initial_state(circuit)

        output = {
            "unitary": circuit_unitary.toarray(),
            "statevector": final_state.toarray().reshape(self._dim),
        }

        # If shots is specified, simulate measurements on the entire register!
        if self.shots is not None:
            meas_distr = np.abs(final_state.toarray().reshape(self._dim)) ** 2

            if not np.isclose(sum(meas_distr), 1.0):
                raise QiskitColdAtomError("Final statevector is not normalized")

            meas_results = self.draw_shots(meas_distr)
            counts_dict = {
                outcome: list(meas_results).count(outcome) for outcome in set(meas_results)
            }
            output["memory"] = meas_results
            output["counts"] = counts_dict

        # return empty memory and counts dictionary if no shots are specified
        else:
            output["memory"] = []
            output["counts"] = {}

        return output

    def to_operators(self, circuit: QuantumCircuit) -> List[SecondQuantizedOp]:
        """
        Convert a circuit to a list of second quantized operators that describe the generators of the
        gates applied to the circuit. The SecondQuantizedOps generating the gates are embedded in the
        larger space corresponding to the entire circuit.

        Args:
            circuit: A quantum circuit with gates described by second quantized generators

        Returns:
            operators: a list of second-quantized operators, one for each applied gate, in the order
            of the gates in the circuit

        Raises:
            QiskitColdAtomError: - If a given gate can not be converted into a second-quantized operator
                         - If a gate is applied after a measurement instruction
                         - If a circuit instruction other than a Gate, measure, load or barrier is given
            NotImplementedError: If ignore_barriers is False
        """
        operators = []
        measured = [False] * circuit.num_qubits

        for inst in circuit.data:
            name = inst[0].name
            qargs = [circuit.qubits.index(qubit) for qubit in inst[1]]

            if name == "measure":
                for idx in qargs:
                    measured[idx] = True

            elif name == "load":
                continue

            elif name == "barrier":
                if self.ignore_barriers:
                    continue
                raise NotImplementedError

            elif isinstance(inst[0], Gate):
                try:
                    second_quantized_op = inst[0].generator
                except AttributeError as attribute_error:
                    raise QiskitColdAtomError(
                        f"Gate {inst[0].name} has no defined generator"
                    ) from attribute_error

                if not isinstance(second_quantized_op, SecondQuantizedOp):
                    raise QiskitColdAtomError(
                        "Gate generator needs to be initialized as qiskit_nature SecondQuantizedOp"
                    )
                for idx in qargs:
                    if measured[idx]:
                        raise QiskitColdAtomError(
                            f"Simulator cannot handle gate {name} after previous measure instruction."
                        )

                if not second_quantized_op.register_length == len(qargs):
                    raise QiskitColdAtomError(
                        f"length of operator labels {second_quantized_op.register_length} must be "
                        f"equal to length of wires {len(qargs)} the gate acts on"
                    )
                operators.append(
                    self._embed_operator(second_quantized_op, circuit.num_qubits, qargs)
                )

            else:
                raise QiskitColdAtomError(f"Unknown instruction {name} applied to circuit")

        return operators

    @abstractmethod
    def get_initial_state(self, circuit: QuantumCircuit) -> csc_matrix:
        """Returns the initial state of the quantum circuit as a sparse column vector."""

    @abstractmethod
    def _embed_operator(
        self, operator: SecondQuantizedOp, num_wires: int, qargs: List[int]
    ) -> SecondQuantizedOp:
        """
        Turning an operator that acts on the wires given in qargs into an operator
        that acts on the entire state space of a circuit. The implementation of the subclasses
        depends on whether the operators use sparse labels (SpinOp) or dense labels (FermionicOp).

        Args:
            operator: SecondQuantizedOp describing the generating Hamiltonian of a gate
            num_wires: number of wires of the space in which to embed the operator
            qargs: The wire indices the gate acts on

        Returns: A SecondQuantizedOp acting on the entire quantum register of the Circuit
        """

    @abstractmethod
    def operator_to_mat(self, operator: SecondQuantizedOp) -> csc_matrix:
        """Turn a SecondQuantizedOp into a sparse matrix."""

    @abstractmethod
    def preprocess_circuit(self, circuit: QuantumCircuit):
        """Pre-process the circuit, e.g. initialize the basis and validate.
        This needs to update ``dim``, i.e. the Hilbert space dimension of the solver."""

    @abstractmethod
    def draw_shots(self, measurement_distribution: List[float]) -> List[str]:
        """
        Simulates shots by drawing from a given distribution of measurement outcomes.
        Assigning the index of each outcome to the occupations of the modes can be
        non-trivial which is why this step needs to be implemented by the subclasses.

        Args:
            measurement_distribution: A list with the probabilities of the different
                measurement outcomes that has the length of the Hilbert space dimension.

        Returns:
            A list of strings encoding the outcome of the individual shots.
        """



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Class that holds a fermionic time-evolution problem."""

from typing import List, Union

from qiskit import QuantumCircuit
from qiskit_nature.operators.second_quantization import FermionicOp

from qiskit_cold_atom.fermions.fermionic_state import FermionicState
from qiskit_cold_atom.fermions.fermionic_basis import FermionicBasis
from qiskit_cold_atom.fermions.fermion_gate_library import FermionicGate
from qiskit_cold_atom.exceptions import QiskitColdAtomError
from qiskit_cold_atom.applications.fermi_hubbard import FermionicLattice


class FermionicEvolutionProblem:
    """
    Problem class corresponding to evaluating an observable of a fermionic system after a time
    evolution under a hamiltonian from an initial state in an occupation number representation.
    """

    def __init__(
        self,
        system: FermionicLattice,
        initial_state: FermionicState,
        evolution_times: Union[float, List[float]],
        observable: FermionicOp,
    ):
        """
        Initialize a fermionic time evolution problem.

        Args:
            system: The fermionic system under which the initial state will evolve.
            initial_state: The fermionic state at time t=0.
            evolution_times: List of times (or single time) after which the observable is measured.
            observable: The observable to measure after the time evolution, given as a FermionicOp.
                        The observable must be diagonal in the fermionic occupation number basis.

        Raises:
            QiskitColdAtomError: - If the sizes of the system, initial state and the observable
                                   do not match.
                                 - If the observables is not diagonal in the fermionic occupation number
                                   basis
        """

        if system.size != initial_state.sites:
            raise QiskitColdAtomError(
                f"The size of the system {system.size} does not match "
                f"the size of the initial state {initial_state.sites}."
            )

        if 2 * system.size != observable.register_length:
            raise QiskitColdAtomError(
                f"The fermionic modes of the system {2*system.size} do not match "
                f"the size of the observable {observable.register_length}."
            )

        # check if matrix is diagonal
        # can later be replaced when the FermionicOp from qiskit-nature has its own .to_matrix() method
        basis = FermionicBasis.from_fermionic_op(observable)
        observable_mat = FermionicGate.operator_to_mat(observable, num_species=1, basis=basis)

        if list(observable_mat.nonzero()[0]) != list(observable_mat.nonzero()[1]):
            raise QiskitColdAtomError(
                "The fermionic observable needs to be diagonal in the computational basis, "
                "as measuring general, non-diagonal observables is not yet implemented for "
                "fermionic backends. This requires non-trivial basis transformations that "
                "are in general difficult to find and depend on the backend's native gate set."
            )

        self._system = system
        self._initial_state = initial_state
        self._evolution_times = evolution_times
        self._observable = observable

    @property
    def system(self) -> FermionicLattice:
        """Return the system of the problem."""
        return self._system

    @property
    def initial_state(self) -> FermionicState:
        """Return the initial state of the system."""
        return self._initial_state

    @property
    def evolution_times(self) -> List[float]:
        """Return the evolution times to simulate."""
        return self._evolution_times

    @property
    def observable(self) -> FermionicOp:
        """Return the observable as a FermionicOp."""
        return self._observable

    def circuits(self, initial_state: QuantumCircuit) -> List[QuantumCircuit]:
        """
        The problem embedded in a quantum circuit.

        Args:
            initial_state: A quantum circuit which corresponds to the initial state for the
                time-evolution problem.

        Return:
            A list of quantum circuits. Circuit :math:`i` is a single instruction which
            corresponds to :math:`exp(-i*H*t_i)` where :math:`t_i` is the time of the
            the ith evolution time.
        """
        circuits = []

        for time in self.evolution_times:
            circ = QuantumCircuit(initial_state.num_qubits)
            circ.compose(initial_state, inplace=True)
            circ.compose(self.system.to_circuit(time), inplace=True)
            circuits.append(circ)

        return circuits



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Gates for fermionic backends."""

from typing import List, Optional, Union
from copy import deepcopy
from scipy.sparse.linalg import expm
from scipy.sparse import csc_matrix
import numpy as np

from qiskit.circuit import Instruction, Gate
from qiskit_nature.operators.second_quantization import FermionicOp

from qiskit_cold_atom.exceptions import QiskitColdAtomError
from qiskit_cold_atom import add_gate
from qiskit_cold_atom.fermions.fermionic_basis import FermionicBasis


class FermionicGate(Gate):
    """Unitary gates for fermionic circuits."""

    def __init__(
        self,
        name: str,
        num_modes: int,
        params: Optional[List] = None,
        label: Optional[str] = None,
        generator: Optional[FermionicOp] = None,
    ) -> None:
        """Create a new fermionic gate.

        Args:
            name: The Qobj name of the gate.
            num_modes: The number of fermionic modes the gate acts on.
            params: A list of parameters.
            label: An optional label for the gate.
            generator: The generating Hamiltonian of the gate unitary given as a FermionicOp
        """

        self._generator = generator

        if params is None:
            params = []

        self.num_modes = num_modes

        super().__init__(name=name, num_qubits=num_modes, params=params, label=label)

    def power(self, exponent: float):
        """Creates a fermionic gate as `gate^exponent`

        Args:
            exponent (float): The exponent with which the gate is exponentiated

        Returns:
            FermionicGate: To which `.generator` is self.generator*exponent.

        Raises:
            QiskitColdAtomError: If the gate has no defined generator.
        """
        if self.generator is None:
            raise QiskitColdAtomError(
                "Gate can not be exponentiated if the gate generator is not defined."
            )
        # the generator of the exponentiated gate is the old generator times the exponent
        exp_generator = exponent * self.generator.simplify()

        exp_params = None if not self.params else [exponent * param for param in self.params]

        exp_label = None if not self.label else self.label + f"^{exponent}"

        return FermionicGate(
            name=self.name + f"^{exponent}",
            num_modes=self.num_modes,
            params=exp_params,
            label=exp_label,
            generator=exp_generator,
        )

    # pylint: disable=arguments-differ
    def to_matrix(self, num_species: int = 1, basis: Optional[FermionicBasis] = None) -> np.ndarray:
        """Return a Numpy.array for the gate unitary matrix. This function will compute :math:`exp(-i H)`
        where :math:`H` is the generator of the gate.

        Args:
            num_species: Number of fermion species which defaults to 1.
            basis: The basis in which to return the matrix. If None is given then the matrix will
                be returned in the full basis.

        Raises:
            QiskitColdAtomError: If the generator of the gate is not hermitian.

        Returns:
            np.array: The array of the gate unitary over the full fock basis, with states ordered like
            00...0, 00...1, ..., 11...0, 11...1
        """
        # This can be simplified when the FermionicOp gets a .to_matrix() method in a future release of
        # qiskit-nature
        if self.generator is None:
            raise QiskitColdAtomError(
                "Matrix of gate can not be computed if the gate generator is not defined."
            )

        generator_mat = self.operator_to_mat(self.generator, num_species, basis)

        if (generator_mat.H - generator_mat).count_nonzero() != 0:
            raise QiskitColdAtomError("generator of unitary gate is not hermitian!")

        unitary_mat = expm(-1j * generator_mat)

        return unitary_mat.toarray()

    @staticmethod
    def operator_to_mat(
        generator: FermionicOp, num_species: int, basis: Optional[FermionicBasis] = None
    ) -> csc_matrix:
        """Compute the matrix representation of the fermion operator.

        Args:
            generator: fermion operator of which to compute the matrix representation.
            num_species: Number of fermion species which defaults to 1.
            basis: The basis in which to return the matrix. If None is given then the matrix will
                be returned in the full basis.

        Returns:
            scipy.sparse matrix of the Hamiltonian.

        Raises:
            QiskitColdAtomError: If the type of the generator is not a FermionicOp.
            QiskitColdAtomError: If the fermion operator does not match the expected shape.
        """

        if not isinstance(generator, FermionicOp):
            raise QiskitColdAtomError(
                f"Expected FermionicOp; got {type(generator).__name__} instead."
            )

        if basis is None:
            basis = FermionicBasis.from_fermionic_op(generator)

        csc_data, csc_col, csc_row = [], [], []

        basis_occupations = basis.get_occupations()

        # loop over all individual terms in the generators
        for term in generator.to_list(display_format="dense"):
            opstring = term[0]
            prefactor = term[1]

            if len(opstring) != num_species * basis.sites:
                raise QiskitColdAtomError(
                    f"Length of operator {opstring} must match the number of "
                    f"modes in the basis {num_species*basis.sites}"
                )

            # loop over all basis states
            for i_basis, occupations in enumerate(basis_occupations):

                new_occupations = deepcopy(occupations)
                mapped_to_zero = (
                    False  # boolean flag to check whether the basis state is mapped to zero
                )
                sign = 1

                # in reverse, loop over all individual fermionic creators/annihilators in the opstring:
                for k, symbol in reversed(list(enumerate(opstring))):

                    if symbol == "I":
                        continue

                    if symbol == "-":
                        # If this mode is not occupied, the action of '-' on this state is zero
                        if occupations[k] == 0:
                            mapped_to_zero = True
                            break
                        sign *= (-1) ** sum(occupations[:k])
                        new_occupations[k] = 0

                    elif symbol == "+":
                        # If this mode is already occupied, the action of '+' on this state is zero
                        if occupations[k] == 1:
                            mapped_to_zero = True
                            break
                        sign *= (-1) ** sum(occupations[:k])
                        new_occupations[k] = 1

                    elif symbol == "N":
                        # If this mode is not occupied, the action of 'N' on this state is zero
                        if occupations[k] == 0:
                            mapped_to_zero = True
                            break

                    elif symbol == "E":
                        # If this mode is occupied, the action of 'E' on this state is zero
                        if occupations[k] == 1:
                            mapped_to_zero = True
                            break

                if not mapped_to_zero:
                    # find the index of the new basis state that the operator strings maps to
                    j_basis = basis_occupations.index(new_occupations)

                    csc_data.append(sign * prefactor)
                    csc_row.append(j_basis)
                    csc_col.append(i_basis)

        return csc_matrix(
            (csc_data, (csc_row, csc_col)),
            shape=(basis.dimension, basis.dimension),
            dtype=complex,
        )

    def control(
        self,
        num_ctrl_qubits: Optional[int] = 1,
        label: Optional[str] = None,
        ctrl_state: Optional[Union[int, str]] = None,
    ):
        """Overwrite control method which is supposed to return a controlled version of the gate.
        This is not applicable in the fermionic setting."""
        raise QiskitColdAtomError("Fermionic gates have no controlled version")

    @property
    def generator(self) -> FermionicOp:
        """The Hamiltonian that generates the unitary of the gate, given as a FermionicOp"""
        return self._generator


class FermiHubbard(FermionicGate):
    r"""
    Global 1D-Fermi-Hubbard dynamic consisting of the hopping, interaction and local phase gates.

    The generating Hamiltonian of the Fermi-Hubbard gate is

    :math:`H = \sum_{i=1,\sigma}^{L-1} - J_i (f^\dagger_{i,\sigma} f_{i+1,\sigma}
    + f^\dagger_{i+1,\sigma} f_{i,\sigma})
    + U \sum_{i=1}^{L} n_{i,\uparrow} n_{i,\downarrow}
    + \sum_{i=1,\sigma}^{L} \mu_i n_{i,\sigma}`

    where :math:`i` indexes the mode, :math:`\sigma` indexes the spin, :math:`L` gives the total number
    of sites, :math:`J_i` are the hopping strengths, :math:`U` is the interaction strength and
    :math:`\mu_i` are the local potentials that lead to local phases.

    **Circuit symbol:**

        .. parsed-literal::

                 ┌──────────────┐
            q_0: ┤   FHubbard   ├
                 └──────────────┘

    """

    def __init__(self, num_modes: int, j: List[float], u: float, mu: List[float], label=None):
        """Initialize a global Fermi-Hubbard gate

        Args:
            num_modes: number of tweezers on which the hopping acts, must be entire quantum register
            j: list of hopping strengths between the tweezer. j[0] gives the strength of hopping
                between wires 0 and 1, j[1] gives the strength of hopping between wires 1 and 2, etc.,
                so len(j) has to be of length num_wires-1
            label: optional
            u: global interaction strength parameter
            mu: list of parameters that tune the local phases, must be on length num_wires
            label: optional

        Raises:
            QiskitColdAtomError:
                - If the given num_modes is not an even integer.
                - If length of j not num_modes/2 - 1.
        """
        if not (isinstance(num_modes, int) and num_modes % 2 == 0):
            raise QiskitColdAtomError("num_modes has to be even integer")
        if not len(j) == (num_modes / 2 - 1):
            raise QiskitColdAtomError("j has to be a list of length num_modes/2 -1")

        param_list = [j, [u], mu]
        params = [item for sublist in param_list for item in sublist]

        super().__init__(
            name="fhubbard",
            num_modes=num_modes,
            params=params,
            label=label,
        )

    def inverse(self):
        """Get the inverse gate by reversing the sign of all gate parameters"""
        j_val, u_val, mu_vals = self.params[0], self.params[1], self.params[2]

        return FermiHubbard(num_modes=self.num_modes, j=-j_val, u=-u_val, mu=-mu_vals)

    @property
    def generator(self) -> FermionicOp:
        """The generating Hamiltonian of the FH Gate."""
        params = [float(param) for param in self.params]
        generators = []
        sites = self.num_modes // 2
        # add generators of hopping term
        if not all(j == 0.0 for j in params[: sites - 1]):
            for i in range(sites - 1):
                generators.append((f"+_{i} -_{i+1}", -1 * params[i]))
                generators.append((f"-_{i} +_{i+1}", params[i]))
                generators.append((f"+_{i+sites} -_{i+sites+1}", -1 * params[i]))
                generators.append((f"-_{i+sites} +_{i+sites+1}", params[i]))
        # add generators of interaction term
        if params[sites - 1] != 0.0:
            for i in range(sites):
                generators.append((f"N_{i} N_{i + sites}", params[sites - 1]))
        # add generators of local phase term
        if not all(muval == 0.0 for muval in params[sites:]):
            for i in range(sites):
                generators.append((f"N_{i}", float(self.params[i + sites])))
                generators.append((f"N_{i+sites}", float(self.params[i + sites])))

        if not generators:
            return FermionicOp("I_0", register_length=self.num_modes)
        else:
            return sum(
                coeff * FermionicOp(label, register_length=self.num_modes)
                for label, coeff in generators
            )


# pylint: disable=invalid-name
@add_gate
def fhubbard(self, j: List[float], u: float, mu: List[float], modes: List[int], label=None):
    """Add the FermiHubbard gate to a QuantumCircuit."""
    return self.append(
        FermiHubbard(num_modes=len(modes), j=j, u=u, mu=mu, label=label), qargs=modes
    )


class Hop(FermionicGate):
    r"""
    Hopping of particles to neighbouring wells due to tunneling.

    The generating Hamiltonian of the hopping gate is

    :math:`H = \sum_{i=1,\sigma}^{L-1} - J_i (f^\dagger_{i,\sigma} f_{i+1,\sigma}
    + f^\dagger_{i+1,\sigma} f_{i,\sigma})`

    where :math:`i` indexes the mode, :math:`\sigma`
    indexes the spin, :math:`L` gives the total number of sites and :math:`J_i` are the hopping strengths
    """

    def __init__(self, num_modes, j: List[float], label=None):
        """
        Initialize hopping gate
        Args:
            num_modes: number of fermionic modes that are connected by the hopping
            (= 2* number of tweezers)
            j: list of hopping strengths between the tweezer. j[0] gives the strength of hopping
            between wires 0 and 1,
            j[1] gives the strength of hopping between wires 1 and 2, etc., so len(j) has to be
            of length num_wires-1
            label: optional
        Raises:
            QiskitColdAtomError: given num_modes not even integer
            QiskitColdAtomError: length of j not num_modes/2 - 1
        """
        if not (isinstance(num_modes, int) and num_modes % 2 == 0):
            raise QiskitColdAtomError("num_modes has to be even integer")
        if not len(j) == (num_modes / 2 - 1):
            raise QiskitColdAtomError("j has to be a list of length num_modes/2 -1")

        super().__init__(name="fhop", num_modes=num_modes, params=j, label=label)

    def inverse(self):
        """Get inverse gate by reversing the sign of all hopping strengths"""
        return Hop(num_modes=self.num_modes, j=self.params)

    @property
    def generator(self) -> FermionicOp:
        """The generating Hamiltonian of the hopping gate."""
        generator = FermiHubbard(
            num_modes=self.num_modes, j=self.params, u=0.0, mu=[0.0]
        ).generator.simplify()

        if generator == 0:
            return FermionicOp("I_0", register_length=self.num_modes)
        else:
            return generator


@add_gate
def fhop(self, j: List[float], modes: List[int], label=None):
    """Add the hopping gate to a QuantumCircuit."""
    return self.append(Hop(num_modes=len(modes), j=j, label=label), qargs=modes)


class Interaction(FermionicGate):
    r"""On-site interaction of particles of opposite spin species on the same site.

    The generating Hamiltonian of the interaction gate is

    :math:`H = U \sum_{i=1}^{L} n_{i,\uparrow} n_{i,\downarrow}`

    where :math:`i` indexes the mode,
    :math:`L` gives the total number of sites and :math:`U` is the interaction strength
    """

    def __init__(self, num_modes: int, u: float, label=None):
        """Initialize interaction gate.

        Args:
            num_modes: number of modes on which the gate acts
            u: global interaction strength parameter
            label: optional

        Raises:
            QiskitColdAtomError: If the number of wires the gate acts on is uneven
        """

        if not num_modes % 2 == 0:
            raise QiskitColdAtomError(f"number of modes must be even, {num_modes} given.")

        super().__init__(
            name="fint",
            num_modes=num_modes,
            params=[u],
            label=label,
        )

    def inverse(self):
        """Get inverse gate by reversing the sign of the interaction parameter"""
        return Interaction(num_modes=self.num_modes, u=-self.params[0])

    @property
    def generator(self) -> FermionicOp:
        """The generating Hamiltonian of the interaction gate."""
        generator = FermiHubbard(
            num_modes=self.num_modes,
            j=[0.0] * (int(self.num_modes / 2) - 1),
            u=self.params[0],
            mu=[0.0],
        ).generator.simplify()

        if generator == 0:
            return FermionicOp("I_0", register_length=self.num_modes)
        else:
            return generator


@add_gate
def fint(self, u: float, modes: List[int], label=None):
    """Add the interaction gate to a QuantumCircuit."""
    return self.append(Interaction(num_modes=len(modes), u=u, label=label), qargs=modes)


class Phase(FermionicGate):
    r"""
    Applying a local phase to individual tweezers through an external potential

    The generating Hamiltonian of the local phase gate is

    :math:`H = \sum_{i=1,\sigma}^{L} \mu_i n_{i,\sigma}`

    where :math:`i` indexes the mode,
    :math:`\sigma` indexes the spin, :math:`L` gives the total number of sites and
    :math:`\mu_i` are the local potentials that lead to local phases
    """

    def __init__(self, num_modes: int, mu: List[float], label=None):
        """
        Initialize a LocalPhase gate
        Args:
            num_modes: number of modes on which the local potential acts
            mu: list of parameters that tune the local phases, must be of length num_modes/2
            label: optional
        Raises:
            QiskitColdAtomError: If the length of mu does not match the given wire count num_modes
        """

        if not num_modes % 2 == 0:
            raise QiskitColdAtomError(f"number of modes must be even, {num_modes} given.")

        if not len(mu) == num_modes / 2:
            raise QiskitColdAtomError(
                f"list of pre-factors {mu} has to be same dimension as the wire count "
                f"of the gate {num_modes}"
            )

        super().__init__(
            name="fphase",
            num_modes=num_modes,
            params=mu,
            label=label,
        )

    def inverse(self):
        """Get inverse gate by reversing the sign of all potentials"""
        return Phase(num_modes=self.num_modes, mu=[-1 * param for param in self.params])

    @property
    def generator(self) -> FermionicOp:
        """The generating Hamiltonian of the local phase gate."""
        generator = FermiHubbard(
            num_modes=self.num_modes,
            j=[0.0] * (int(self.num_modes / 2) - 1),
            u=0.0,
            mu=self.params,
        ).generator.simplify()

        if generator == 0:
            return FermionicOp("I_0", register_length=self.num_modes)
        else:
            return generator


@add_gate
def fphase(self, mu: List[float], modes: List[int], label=None):
    """Add the local phase gate to a QuantumCircuit."""
    return self.append(Phase(num_modes=len(modes), mu=mu, label=label), qargs=modes)


class FRXGate(FermionicGate):
    r"""X-rotation between the spin-up and spin-down state at one tweezer site.

    The generating Hamiltonian of the FermionRx gate is

    :math:`H = \phi (f^\dagger_{x,\uparrow} f_{x,\downarrow} + f^\dagger_{x,\downarrow} f_{x,\uparrow})`

    where :math:`x` is the index of the mode and :math:`\phi` is the free gate parameter
    """

    def __init__(self, phi: float, label=None):
        """Initialize a FermionRX gate

        Args:
            phi: angle of the rotation
            label: optional
        """

        super().__init__(name="frx", num_modes=2, params=[phi], label=label)

    def inverse(self):
        """Get inverse gate by inverting the sign of the rotation angle"""
        return FRXGate(-self.params[0])

    @property
    def generator(self) -> FermionicOp:
        """The generating Hamiltonian of the FermionRX gate."""
        op = float(self.params[0]) * FermionicOp("+_0 -_1", register_length=2) - float(
            self.params[0]
        ) * FermionicOp("-_0 +_1", register_length=2)
        return op


@add_gate
def frx(self, phi: float, wires: List[int]):
    """Add the FermionRX gate to a QuantumCircuit."""
    return self.append(FRXGate(phi=phi), qargs=wires)


class FRYGate(FermionicGate):
    r"""Y-rotation between the spin-up and spin-down state at one tweezer site.

    The generating Hamiltonian of the FermionRy gate is

    :math:`H = \phi (f^\dagger_{x,\uparrow} f_{x,\uparrow} - f^\dagger_{x,\downarrow} f_{x,\downarrow})`

    where :math:`x` is the index of the mode and :math:`\phi` is the free gate parameter
    """

    def __init__(self, phi, label=None):
        """
        Initialize a FermionRY gate
        Args:
            phi: angle of the rotation
            label: optional
        """
        super().__init__(name="fry", num_modes=2, params=[phi], label=label)

    def inverse(self):
        """Get inverse gate by inverting the sign of the rotation angle"""
        return FRYGate(-self.params[0])

    @property
    def generator(self) -> FermionicOp:
        """The generating Hamiltonian of the FermionRY gate."""
        op = -1j * float(self.params[0]) * FermionicOp("+_0 -_1", register_length=2) - 1j * float(
            self.params[0]
        ) * FermionicOp("-_0 +_1", register_length=2)
        return op


@add_gate
def fry(self, phi: float, wires: List[int]):
    """Add the FermionRY gate to a QuantumCircuit."""
    return self.append(FRYGate(phi=phi), qargs=wires)


class FRZGate(FermionicGate):
    r"""Z-rotation between the spin-up and spin-down state at one tweezer site.

    The generating Hamiltonian of the FermionRz gate is

    :math:`H = i \phi (f^\dagger_{x,\downarrow} f_{x,\uparrow}-f^\dagger_{x,\uparrow} f_{x,\downarrow})`

    where :math:`x` is the index of the mode and :math:`\phi` is the free gate parameter.
    """

    def __init__(self, phi, label=None):
        """
        Initialize a FermionRZ gate
        Args:
            phi: angle of the rotation
            label: optional
        """
        super().__init__(name="frz", num_modes=2, params=[phi], label=label)

    def inverse(self):
        """Get inverse gate by inverting the sign of the rotation angle"""
        return FRZGate(-self.params[0])

    @property
    def generator(self) -> FermionicOp:
        """The generating Hamiltonian of the FermionRZ gate."""
        op = float(self.params[0]) * FermionicOp("N_0", register_length=2) - float(
            self.params[0]
        ) * FermionicOp("N_1", register_length=2)
        return op


@add_gate
def frz(self, phi: float, wires: List[int]):
    """Add the FermionRZ gate to a QuantumCircuit."""
    return self.append(FRZGate(phi=phi), qargs=wires)


class LoadFermions(Instruction):
    """
    LoadFermions places a particle in an empty fermionic mode.

    **Circuit symbol:**

    .. parsed-literal::

             ┌──────┐
        q_0: ┤ Load ├
             └──────┘
    """

    def __init__(self):
        """Initialise new load instruction."""
        super().__init__(name="load", num_qubits=1, num_clbits=0, params=[])


@add_gate
def fload(self, wire):
    """Add the load fermion gate to a QuantumCircuit."""
    return self.append(LoadFermions(), [wire], [])



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Backend to emulate collective spin experiments."""

from typing import Optional

from qiskit.providers import ProviderV1 as Provider

from qiskit_cold_atom.spins.spin_simulator_backend import SpinSimulator


class CollectiveSpinSimulator(SpinSimulator):
    """Emulator backend of a spin device with collective spins of trapped BECs in optical tweezers.

    This backend will define a configuration with a set of supported gates and a line coupling map.
    At instantiation users can chose the length of the coupling map which by default is three. This
    backend is intended to be a realistic representation of a collective spin experiment with BECs.
    For a general spin simulator backend, use the SpinSimulator base class.
    """

    def __init__(self, n_tweezers: int = 3, provider: Optional[Provider] = None):
        """Create a new collective spin simulator backend.

        Args:
            n_tweezers: The number of optical tweezers.
            provider: The provider to which the backend may be added.
        """

        single_spin_coupl = [[i] for i in range(n_tweezers)]
        two_spin_coupl = [[i, i + 1] for i in range(n_tweezers - 1)] + [
            [i + 1, i] for i in range(n_tweezers - 1)
        ]

        configuration = {
            "backend_name": "collective_spin_simulator",
            "cold_atom_type": "spin",
            "backend_version": "0.0.1",
            "atomic_species": ["na"],
            "simulator": True,
            "local": True,
            "coupling_map": None,
            "description": "simulator of a collective hardware. Each wire in the circuit denotes the "
            "orientation of one collective spin in the hardware",
            "basis_gates": ["rlx", "rly", "rlz", "rlz2", "OAT", "rlzlz", "rlxly"],
            "memory": True,
            "n_qubits": 3,
            "conditional": False,
            "max_shots": 1e6,
            "max_experiments": 10,
            "open_pulse": False,
            "gates": [
                {
                    "coupling_map": single_spin_coupl,
                    "description": "local rotation of the coherent spin around x",
                    "name": "rlx",
                    "parameters": ["omega"],
                    "qasm_def": "gate rlx(omega) {}",
                },
                {
                    "coupling_map": single_spin_coupl,
                    "description": "local rotation of the coherent spin around y",
                    "name": "rly",
                    "parameters": ["delta"],
                    "qasm_def": "gate rly(delta) {}",
                },
                {
                    "coupling_map": single_spin_coupl,
                    "description": "local rotation of the coherent spin around z",
                    "name": "rlz",
                    "parameters": ["delta"],
                    "qasm_def": "gate rlz(delta) {}",
                },
                {
                    "coupling_map": single_spin_coupl,
                    "description": "local spin evolution generated by one-axis "
                    "twisting Hamiltonian Lz^2",
                    "name": "rlz2",
                    "parameters": ["chi"],
                    "qasm_def": "gate rlz2(chi) {}",
                },
                {
                    "coupling_map": single_spin_coupl,
                    "description": "local spin evolution generated by (chi*Lz^2 + "
                    "delta*Lz + omega*Lx)",
                    "name": "oat",
                    "parameters": ["chi", "delta", "omega"],
                    "qasm_def": "gate rot(chi, delta, omega) {}",
                },
                {
                    "coupling_map": two_spin_coupl,
                    "description": "ZZ-interaction coupling two collective spins",
                    "name": "rlzlz",
                    "parameters": ["gamma"],
                    "qasm_def": "gate rlzlz(gamma) {}",
                },
                {
                    "coupling_map": two_spin_coupl,
                    "description": "LXLY-interaction coupling two collective spins",
                    "name": "rlxly",
                    "parameters": ["gamma"],
                    "qasm_def": "gate rlxly(gamma) {}",
                },
            ],
            "supported_instructions": [
                "rlz",
                "rlx",
                "rly",
                "rlz2",
                "OAT",
                "rlzlz",
                "rlxly",
                "measure",
                "barrier",
            ],
        }

        super().__init__(config_dict=configuration, provider=provider)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Spin circuit solver tests"""

import numpy as np

from qiskit import QuantumCircuit
from qiskit.test import QiskitTestCase
from qiskit_nature.operators.second_quantization import FermionicOp, SpinOp
from qiskit_cold_atom.spins.spin_circuit_solver import SpinCircuitSolver
from qiskit_cold_atom.exceptions import QiskitColdAtomError


class TestSpinCircuitSolver(QiskitTestCase):
    """class to test the SpinCircuitSolver class."""

    def setUp(self):
        super().setUp()
        # Set up the simulator
        self.solver = SpinCircuitSolver(spin=3 / 2)

    def test_spin_solver_initialization(self):
        """test constructor of SpinCircuitSolver"""
        with self.assertRaises(QiskitColdAtomError):
            SpinCircuitSolver(spin=2 / 3)

    def test_get_initial_state(self):
        """test initialization of the state for the simulation"""
        circ = QuantumCircuit(1)
        init_state = self.solver.get_initial_state(circ)
        target = np.array([1, 0, 0, 0])
        self.assertTrue(np.alltrue(init_state.toarray().T == target))

    def test_embed_operator(self):
        """test embedding of an operator"""
        fer_op = FermionicOp("+-")
        # define a spin operator that has terms with different prefactors, support and power
        spin_op = SpinOp("+-") + 2 * SpinOp("X_0^2", register_length=2)
        num_wires = 4
        qargs = [1, 3]
        qargs_wrong = [0, 1, 3]

        with self.subTest("check operator type"):
            with self.assertRaises(QiskitColdAtomError):
                self.solver._embed_operator(fer_op, num_wires, qargs)

        with self.subTest("check operator wiring"):
            with self.assertRaises(QiskitColdAtomError):
                self.solver._embed_operator(spin_op, num_wires, qargs_wrong)

        with self.subTest("operator embedding"):
            embedded_op = self.solver._embed_operator(spin_op, num_wires, qargs)
            target_op = SpinOp([("+_1 -_3", 1.0), ("X_1^2", 2.0)], spin=3 / 2, register_length=4)
            self.assertTrue(
                set(embedded_op.simplify().to_list()) == set(target_op.simplify().to_list())
            )

    def test_preprocess_circuit(self):
        """test whether preprocessing of the circuit correctly sets the dimension"""
        circ = QuantumCircuit(2)
        self.solver.preprocess_circuit(circ)
        self.assertEqual(self.solver.dim, 4**2)

    def test_draw_shots(self):
        """test drawing of the shots from a measurement distribution"""
        n_spins = 5
        circ = QuantumCircuit(n_spins)
        circ.rly(np.pi / 2, [3])
        self.solver.preprocess_circuit(circ)
        dim = int((2 * self.solver.spin + 1) ** n_spins)

        with self.subTest("check missing shot number"):
            # error because the number of shots is not specified
            with self.assertRaises(QiskitColdAtomError):
                self.solver.draw_shots(np.ones(dim) / dim)

        self.solver.shots = 3

        with self.subTest("check match of dimensions"):
            # error because there is a mismatch in the dimension
            with self.assertRaises(QiskitColdAtomError):
                self.solver.draw_shots(np.ones(dim - 1) / (dim - 1))

        with self.subTest("formatting of measurement outcomes"):
            meas_distr = np.abs(self.solver(circ)["statevector"]) ** 2
            self.solver.seed = 45
            outcomes = self.solver.draw_shots(meas_distr)
            # Note the second index changing with the gate being applied to wire [3] of 5
            self.assertEqual(outcomes, ["0 3 0 0 0", "0 2 0 0 0", "0 1 0 0 0"])

    def test_to_operators(self):
        """test the to_operators method inherited form BaseCircuitSolver"""

        test_circ = QuantumCircuit(2)
        test_circ.rlx(0.5, [0, 1])
        test_circ.rlz2(0.25, 1)
        test_circ.measure_all()

        with self.subTest("test ignore barriers"):
            self.solver.ignore_barriers = False
            with self.assertRaises(NotImplementedError):
                self.solver.to_operators(test_circ)
            self.solver.ignore_barriers = True

        with self.subTest("check for gate generators"):
            qubit_circ = QuantumCircuit(1)
            qubit_circ.h(0)
            with self.assertRaises(QiskitColdAtomError):
                self.solver.to_operators(qubit_circ)

        with self.subTest("gate after previous measurement instruction"):
            meas_circ = QuantumCircuit(2)
            meas_circ.measure_all()
            meas_circ.rlx(0.5, 0)
            with self.assertRaises(QiskitColdAtomError):
                self.solver.to_operators(meas_circ)

        with self.subTest("check returned operators"):
            operators = self.solver.to_operators(test_circ)
            target = [
                SpinOp([("X_0", (0.5 + 0j))], spin=3 / 2, register_length=2),
                SpinOp([("X_1", (0.5 + 0j))], spin=3 / 2, register_length=2),
                SpinOp([("Z_1^2", (0.25 + 0j))], spin=3 / 2, register_length=2),
            ]

            for i, op in enumerate(operators):
                self.assertEqual(set(op.simplify().to_list()), set(target[i].simplify().to_list()))

    def test_call_method(self):
        """test the call method inherited form BaseCircuitSolver that simulates a circuit"""

        test_circ = QuantumCircuit(1)
        test_circ.rlx(np.pi / 2, 0)
        test_circ.measure_all()

        with self.subTest("running the circuit"):
            self.solver.shots = 5
            self.solver.seed = 45
            simulation = self.solver(test_circ)

            self.assertEqual(simulation["memory"], ["3", "2", "1", "0", "1"])
            self.assertEqual(simulation["counts"], {"0": 1, "1": 2, "2": 1, "3": 1})
            self.assertTrue(
                np.allclose(
                    simulation["statevector"],
                    np.array(
                        [
                            np.sqrt(1 / 8),
                            -1j * np.sqrt(3 / 8),
                            -np.sqrt(3 / 8),
                            1j * np.sqrt(1 / 8),
                        ]
                    ),
                )
            )
            self.assertTrue(
                np.allclose(
                    simulation["unitary"],
                    np.array(
                        [
                            [
                                np.sqrt(1 / 8),
                                -1j * np.sqrt(3 / 8),
                                -np.sqrt(3 / 8),
                                1j * np.sqrt(1 / 8),
                            ],
                            [
                                -1j * np.sqrt(3 / 8),
                                -np.sqrt(1 / 8),
                                -1j * np.sqrt(1 / 8),
                                -np.sqrt(3 / 8),
                            ],
                            [
                                -np.sqrt(3 / 8),
                                -1j * np.sqrt(1 / 8),
                                -np.sqrt(1 / 8),
                                -1j * np.sqrt(3 / 8),
                            ],
                            [
                                1j * np.sqrt(1 / 8),
                                -np.sqrt(3 / 8),
                                -1j * np.sqrt(3 / 8),
                                np.sqrt(1 / 8),
                            ],
                        ]
                    ),
                )
            )

        with self.subTest("check for maximum dimension"):
            self.solver.max_dimension = 3
            with self.assertRaises(QiskitColdAtomError):
                self.solver(test_circ)
            self.solver.max_dimension = 100

        with self.subTest("check if shots are specified"):
            self.solver.shots = None
            simulation = self.solver(test_circ)
            self.assertEqual(simulation["memory"], [])
            self.assertEqual(simulation["counts"], {})
            self.solver.shots = 5

        multiple_wire_circ = QuantumCircuit(2)
        multiple_wire_circ.rlx(np.pi / 2, [0])
        multiple_wire_circ.rlz(-np.pi / 2, [0, 1])

        with self.subTest("formatting of multiple wires"):
            self.solver.seed = 45
            simulation = self.solver(multiple_wire_circ)
            self.assertTrue(simulation["memory"], ["0 3", "0 2", "0 1", "0 0", "0 1"])
            self.assertTrue(simulation["counts"], {"0 2": 1, "0 0": 1, "0 3": 1, "0 1": 2})

        with self.subTest("check equivalence to qubits for spin-1/2"):
            from qiskit_aer import AerSimulator

            qubit_circ = QuantumCircuit(2)
            qubit_circ.rx(np.pi / 2, [0])
            qubit_circ.rz(-np.pi / 2, [0, 1])
            qubit_circ.save_unitary()
            qubit_backend = AerSimulator()
            job = qubit_backend.run(qubit_circ)
            qubit_unitary = job.result().get_unitary()

            spin_half_solver = SpinCircuitSolver(spin=1 / 2)
            simulation = spin_half_solver(multiple_wire_circ)
            spin_unitary = simulation["unitary"]
            # Switch some axes in the spin simulator unitary because the basis of qiskit_nature.SpinOp
            # uses an ordering of the states (00, 10, 01, 11) that is different from qiskit Aer which
            # uses (00, 01, 10, 11)
            spin_unitary[[1, 2]] = spin_unitary[[2, 1]]
            spin_unitary[:, [1, 2]] = spin_unitary[:, [2, 1]]

            self.assertTrue(np.allclose(qubit_unitary, spin_unitary))



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Module to simulate spin circuits."""

import math
from typing import List, Union, Optional
from fractions import Fraction
import numpy as np
from scipy.sparse import csc_matrix
from qiskit import QuantumCircuit
from qiskit_nature.operators.second_quantization import SpinOp
from qiskit_cold_atom.base_circuit_solver import BaseCircuitSolver
from qiskit_cold_atom.exceptions import QiskitColdAtomError


class SpinCircuitSolver(BaseCircuitSolver):
    """Performs numerical simulations of spin systems by exactly computing the time
    evolution under unitary operations generated by exponentiating spin Hamiltonians."""

    def __init__(
        self,
        spin: Union[float, Fraction] = Fraction(1, 2),
        shots: Optional[int] = None,
        seed: Optional[int] = None,
    ):
        """
        Initialize a spin circuit solver.

        Args:
            spin: The length of the spin of each wire in the circuit.
            shots: Amount of shots for the measurement simulation;
                   if not None, measurements are performed.
            seed: The seed for the RNG for the measurement simulation.

        Raises:
            QiskitColdAtomError: if the spin is not a positive integer or half-integer.
        """

        self.spin = Fraction(spin)
        if self.spin.denominator not in (1, 2):
            raise QiskitColdAtomError(
                f"spin must be a positive half-integer (integer or half-odd-integer), "
                f"not {self.spin}."
            )

        super().__init__(shots=shots, seed=seed)

    def get_initial_state(self, circuit: QuantumCircuit) -> csc_matrix:
        """
        Return the initial state as a sparse column vector.

        Args:
            circuit: A circuit that tells us the dimension of the initial state to return.

        Returns:
            initial state: A sparse column vector of the initial state.
        """

        dim = int((2 * self.spin + 1) ** circuit.num_qubits)

        initial_state = csc_matrix(([1 + 0j], ([0], [0])), shape=(dim, 1), dtype=complex)
        return initial_state

    def _embed_operator(self, operator: SpinOp, num_wires: int, qargs: List[int]) -> SpinOp:
        """
        Turning a SpinOp operator that acts onto the wires given in qargs into an operator
        that acts on the entire register of the circuit by manipulating the indices of the
        sparse labels of the SpinOps.

        Args:
            operator: SpinOp describing the generating Hamiltonian of a gate
            num_wires: The total number of wires in which the operator should be embedded into
            qargs: The wire indices the gate acts on

        Returns:
            A SpinOp acting on the entire quantum register of the Circuit

        Raises:
            QiskitColdAtomError: - If the given operator is not a SpinOp
                                 - If the size of the operator does not match the given qargs
        """
        if not isinstance(operator, SpinOp):
            raise QiskitColdAtomError(f"Expected SpinOp; got {type(operator).__name__} instead")

        if operator.register_length != len(qargs):
            raise QiskitColdAtomError(
                f"operator size {operator.register_length} does not match qargs {qargs} of the gates."
            )

        embedded_op_list = []
        for label, factor in operator.to_list():
            old_labels = label.split()
            new_labels = [term[:2] + str(qargs[int(term[2])]) + term[3:] for term in old_labels]
            embedded_op_list.append((" ".join(map(str, new_labels)), factor))

        return SpinOp(embedded_op_list, spin=self.spin, register_length=num_wires)

    def operator_to_mat(self, operator: SpinOp) -> csc_matrix:
        """
        Convert a SpinOp describing a gate generator to a sparse matrix.

        Args:
            operator: spin operator of which to compute the matrix representation

        Returns:
            scipy.sparse matrix of the Hamiltonian
        """
        return csc_matrix(operator.to_matrix())

    def preprocess_circuit(self, circuit: QuantumCircuit):
        r"""
        Compute the Hilbert space dimension of the given quantum circuit as :math:`(2S+1)^N`
        where :math:`S` is the length of the spin and :math:`N` is the number of spins in
        the quantum circuit.

        Args:
            circuit: The circuit to pre-process.
        """
        self._dim = int((2 * self.spin + 1) ** circuit.num_qubits)

    def draw_shots(self, measurement_distribution: List[float]) -> List[str]:
        r"""A helper function to draw counts from a given distribution of measurement outcomes.

        Args:
            measurement_distribution: List of probabilities of the individual measurement outcomes.

        Returns:
            outcome_memory: A list of individual measurement results, e.g. ["12 3 4", "0 4 9", ...]
            The outcome of each shot is denoted by a space-delimited string "a1 a2 a3 ..." where
            :math:`a_i` is the measured level of the spin with possible values ranging from 0 to 2S
            The :math:`a_i` are in reverse order of the spins of the register to comply with qiskit's
            little endian convention.

        Raises:
            QiskitColdAtomError:
                - If the length of the given probabilities does not math the expected Hilbert
                    space dimension.
                - If the dimension is not a power of the spin length of the solver.
                - If the number of shots self.shots has not been specified.
        """

        meas_dim = len(measurement_distribution)

        if meas_dim != self.dim:
            raise QiskitColdAtomError(
                f"Dimension of the measurement probabilities {meas_dim} does not "
                f"match the dimension expected by the solver, {self.dim}"
            )

        if self.shots is None:
            raise QiskitColdAtomError(
                "The number of shots has to be set before drawing measurements"
            )

        # Draw measurements as the indices of the basis states:
        meas_results = np.random.choice(range(meas_dim), self.shots, p=measurement_distribution)
        base = int(2 * self.spin + 1)
        num_wires = math.log(meas_dim, base)

        if num_wires.is_integer():
            num_wires = int(num_wires)
        else:
            raise QiskitColdAtomError(
                "The length of given measurement distribution it not compatible with "
                "the spin-length of the solver."
            )

        outcome_memory = []
        for meas_idx in meas_results:
            digits = [0] * num_wires
            for i in range(num_wires):
                digits[i] = meas_idx % base
                meas_idx //= base

            outcome_memory.append(" ".join(map(str, digits)))

        return outcome_memory



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Module to study applications using cold atomic setups. Users of the qiskit_cold_atoms.applications
module can create Fermionic time-evolution problems to simulate their dynamics on simulator
backends or cold-atom based hardware that natively supports the Hamiltonian of the fermionic
problem.

For lattice-based problems, users must create subclasses of the :class:`FermionicLattice` class.
An example of which is the one-dimensional :class:`FermiHubbard1D` which describes a
one-dimensional Fermi-Hubbard lattice with spin-up and spin-down particles.

.. autosummary::
   :toctree: ../stubs/

    FermionicLattice
    FermiHubbard1D
    FermionicEvolutionProblem
    TimeEvolutionSolver

"""

from .fermi_hubbard import FermionicLattice, FermiHubbard1D
from .fermionic_evolution_problem import FermionicEvolutionProblem
from .time_evolution_solver import TimeEvolutionSolver



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""General Fermion simulator backend."""

from typing import Union, List, Dict, Any, Optional
import uuid
import warnings
import time
import datetime

from qiskit.providers.models import BackendConfiguration
from qiskit.providers import Options
from qiskit import QuantumCircuit
from qiskit.result import Result
from qiskit_aer import AerJob

from qiskit_cold_atom.fermions.fermion_circuit_solver import FermionCircuitSolver
from qiskit_cold_atom.fermions.base_fermion_backend import BaseFermionBackend
from qiskit_cold_atom.circuit_tools import CircuitTools


class FermionSimulator(BaseFermionBackend):
    """A simulator to simulate general fermionic circuits.

    This general fermion simulator backend simulates fermionic circuits with gates that have
    generators described by fermionic Hamiltonians. It computes the statevector and unitary
    of a circuit and simulates measurements.
    """

    _DEFAULT_CONFIGURATION = {
        "backend_name": "fermion_simulator",
        "backend_version": "0.0.1",
        "n_qubits": 20,
        "basis_gates": None,
        "gates": [],
        "local": False,
        "simulator": True,
        "conditional": False,
        "open_pulse": False,
        "memory": True,
        "max_shots": 1e5,
        "coupling_map": None,
        "description": "a base simulator for fermionic circuits. Instead of qubits, each wire represents"
        " a single fermionic mode",
        "supported_instructions": None,
    }

    def __init__(self, config_dict: Dict[str, Any] = None, provider=None):
        """Initializing the backend from a configuration dictionary"""

        if config_dict is None:
            config_dict = self._DEFAULT_CONFIGURATION

        super().__init__(
            configuration=BackendConfiguration.from_dict(config_dict), provider=provider
        )

    @classmethod
    def _default_options(cls):
        return Options(shots=1)

    def _execute(self, data: Dict[str, Any], job_id: str = ""):
        """Helper function to execute a job. The circuit and all relevant parameters are given in the
        data dict. Performs validation checks on the received circuits and utilizes
        the FermionCircuitSolver to perform the numerical simulations.

         Args:
            data: Data dictionary that that contains the experiments to simulate, given in the shape:
                data = {
                    "num_species": int,
                    "shots": int,
                    "seed": int,
                    "experiments": Dict[str, QuantumCircuit],
                }
            job_id: The job id assigned by the run method

        Returns:
            result: A qiskit job result.
        """
        # Start timer
        start = time.time()

        output = {"results": []}

        num_species = data["num_species"]
        shots = data["shots"]
        seed = data["seed"]

        solver = FermionCircuitSolver(num_species=num_species, shots=shots, seed=seed)

        for exp_i, exp_name in enumerate(data["experiments"]):
            experiment = data["experiments"][exp_name]
            circuit = experiment["circuit"]

            # perform compatibility checks with the backend configuration in case gates and supported
            # instructions are constrained by the backend's configuration
            if self.configuration().gates and self.configuration().supported_instructions:
                CircuitTools.validate_circuits(circuits=circuit, backend=self, shots=shots)

            # check whether all wires are measured
            measured_wires = []

            for inst in circuit.data:
                name = inst[0].name

                if name == "measure":

                    for wire in inst[1]:
                        index = circuit.qubits.index(wire)
                        if index in measured_wires:
                            warnings.warn(
                                f"Wire {index} has already been measured, "
                                f"second measurement is ignored"
                            )
                        else:
                            measured_wires.append(index)

            if measured_wires and len(measured_wires) != len(circuit.qubits):
                warnings.warn(
                    f"Number of wires in circuit ({len(circuit.qubits)}) exceeds number of wires "
                    + f" with assigned measurement instructions ({len(measured_wires)}). "
                    + "This simulator backend only supports measurement of the entire quantum register "
                    "which will instead be performed."
                )

            # If there are no measurements, set shots to None
            if not measured_wires:
                solver.shots = None

            simulation_result = solver(circuit)

            output["results"].append(
                {
                    "header": {"name": exp_name, "random_seed": seed},
                    "shots": shots,
                    "status": "DONE",
                    "success": True,
                }
            )
            # add the simulation result at the correct place in the result dictionary
            output["results"][exp_i]["data"] = simulation_result

        output["job_id"] = job_id
        output["date"] = datetime.datetime.now().isoformat()
        output["backend_name"] = self.name()
        output["backend_version"] = self.configuration().backend_version
        output["time_taken"] = time.time() - start
        output["success"] = True
        output["qobj_id"] = None

        return Result.from_dict(output)

    # pylint: disable=arguments-differ, unused-argument
    def run(
        self,
        circuits: Union[QuantumCircuit, List[QuantumCircuit]],
        shots: int = 1000,
        seed: Optional[int] = None,
        num_species: int = 1,
        **run_kwargs,
    ) -> AerJob:
        """
        Method to run circuits on the backend.

        Args:
            circuits: QuantumCircuit applying fermionic gates to run on the backend
            shots: Number of measurement shots taken in case the circuit has measure instructions
            seed: seed for the random number generator of the measurement simulation
            num_species: number of different fermionic species described by the circuits
            run_kwargs: Additional keyword arguments that might be passed down when calling
                qiskit.execute() which will have no effect on this backend.

        Returns:
            aer_job: a job object containing the result of the simulation
        """

        if isinstance(circuits, QuantumCircuit):
            circuits = [circuits]

        data = {
            "num_species": num_species,
            "shots": shots,
            "seed": seed,
            "experiments": {},
        }

        for idx, circuit in enumerate(circuits):
            data["experiments"][f"experiment_{idx}"] = {
                "circuit": circuit,
            }

        job_id = str(uuid.uuid4())
        aer_job = AerJob(self, job_id, self._execute, data)
        aer_job.submit()
        return aer_job

    @staticmethod
    def get_basis(circuit: QuantumCircuit, num_species: int = 1):
        """Get the basis of fermionic states in occupation number representation for the simulation
        of a given quantum circuit.

        Args:
            circuit: A quantum circuit using Fermionic Gates.
            num_species: Number of different fermionic species described by the circuit.

        Returns:
            basis: the fermionic basis in which the simulation of the circuit is performed.
        """
        solver = FermionCircuitSolver(num_species=num_species)
        solver.preprocess_circuit(circuit)
        basis = solver.basis

        return basis

    def draw(self, qc: QuantumCircuit, **draw_options):
        """Draw the circuit by defaulting to the draw method of QuantumCircuit.

        Note that in the future this method may be modified and tailored to fermion
        quantum circuits.

        Args:
            qc: The quantum circuit to draw.
            draw_options: Key word arguments for the drawing of circuits.
        """
        qc.draw(**draw_options)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Module to convert cold atom circuits to dictionaries that can be sent to backends."""

from typing import List, Union, Optional
from enum import Enum

from qiskit import QuantumCircuit
from qiskit.providers import BackendV1 as Backend
from qiskit_cold_atom.exceptions import QiskitColdAtomError


class WireOrder(str, Enum):
    """The possible wire orderings for cold atomic circuits.

    For example, a sequential register [0, 1, 2, 3, 4, 5] with two species implies that wires 0, 1, 2
    are of the same type while an interleaved ordering implies that wires 0, 2, and 4 are of the
    same type.
    """

    SEQUENTIAL = "sequential"
    INTERLEAVED = "interleaved"


class CircuitTools:
    """A class to provide tooling for cold-atomic circuits.

    Since all methods are class methods this class does not need to be instantiated. This
    class groups tools for cold atomic circuits. It also makes clear the ordering of the
    fermionic wires that qiskit works with.
    """

    # Qiskit for fermions works with a sequential register definition. For fermionic
    # modes with more than on species the circuits will have a corresponding number of
    # sequential registers with the same length. For example, a three site system with two
    # species will have two sequential registers with three wires each. Other packages may
    # use an "interleaved" wire order.
    __wire_order__ = WireOrder("sequential")

    @classmethod
    def validate_circuits(
        cls,
        circuits: Union[List[QuantumCircuit], QuantumCircuit],
        backend: Backend,
        shots: Optional[int] = None,
        convert_wires: bool = True,
    ) -> None:
        """
        Performs validity checks on circuits against the configuration of the backends. This checks
        whether all applied instructions in the circuit are accepted by the backend and whether the
        applied gates comply with their respective coupling maps.

        Args:
            circuits: The circuits that need to be run.
            backend: The backend on which the circuit should be run.
            shots: The number of shots for each circuit.
            convert_wires: If True, the circuits are converted to the wiring convention of the backend.

        Raises:
            QiskitColdAtomError: If the maximum shot number specified by the backend is exceeded.
            QiskitColdAtomError: If the backend does not support an instruction in the circuit.
            QiskitColdAtomError: If the width of the circuit is too large.
            QiskitColdAtomError: If the circuit has unbound parameters.
        """

        if isinstance(circuits, QuantumCircuit):
            circuits = [circuits]

        # check for number of experiments allowed by the backend
        if backend.configuration().max_experiments:
            max_circuits = backend.configuration().max_experiments
            if len(circuits) > max_circuits:
                raise QiskitColdAtomError(
                    f"{backend.name()} allows for max. {max_circuits} different circuits; "
                    f"but {len(circuits)} circuits were given"
                )
        # check for number of individual shots allowed by the backend
        if backend.configuration().max_shots and shots:
            max_shots = backend.configuration().max_shots
            if shots > max_shots:
                raise QiskitColdAtomError(
                    f"{backend.name()} allows for max. {max_shots} shots per circuit; "
                    f"{shots} shots were requested"
                )

        config_dict = backend.configuration().to_dict()

        for circuit in circuits:

            try:
                native_gates = {
                    gate.name: gate.coupling_map for gate in backend.configuration().gates
                }
                native_instructions = backend.configuration().supported_instructions
            except NameError as name_error:
                raise QiskitColdAtomError(
                    "backend needs to be initialized with config file first"
                ) from name_error

            if circuit.num_qubits > backend.configuration().num_qubits:
                raise QiskitColdAtomError(
                    f"{backend.name()} supports circuits with up to "
                    f"{backend.configuration().num_qubits} wires, but"
                    f"{circuit.num_qubits} wires were given."
                )

            # If num_species is specified by the backend, the wires describe different atomic species
            # and the circuit's wire count must be a multiple of the number of species.
            num_species = None
            wire_order = None
            if "num_species" in config_dict and convert_wires:
                num_species = backend.configuration().num_species
                if "wire_order" in config_dict:
                    wire_order = WireOrder(backend.configuration().wire_order)
                else:
                    wire_order = cls.__wire_order__

                if num_species > 1 and circuit.num_qubits % num_species:
                    raise QiskitColdAtomError(
                        f"{backend.name()} requires circuits to be submitted with a multiple of "
                        f"{num_species} wires, but {circuit.num_qubits} wires were given."
                    )

            for inst in circuit.data:
                # get the correct wire indices of the instruction with respect
                # to the total index of the qubit objects in the circuit
                wires = [circuit.qubits.index(qubit) for qubit in inst[1]]

                for param in inst[0].params:
                    try:
                        float(param)
                    except TypeError as type_error:
                        raise QiskitColdAtomError(
                            "Cannot run circuit with unbound parameters."
                        ) from type_error

                # check if instruction is supported by the backend
                name = inst[0].name
                if name not in native_instructions:
                    raise QiskitColdAtomError(f"{backend.name()} does not support {name}")

                # for the gates, check whether coupling map fits
                if name in native_gates:
                    couplings = native_gates[name]

                    if num_species and convert_wires:
                        wires = cls.convert_wire_order(
                            wires,
                            convention_from=cls.__wire_order__,
                            convention_to=wire_order,
                            num_species=num_species,
                            num_sites=circuit.num_qubits // num_species,
                            sort=True,
                        )

                    if wires not in couplings:
                        raise QiskitColdAtomError(
                            f"coupling {wires} not supported for gate "
                            f"{name} on {backend.name()}; possible couplings: {couplings}"
                        )

    @classmethod
    def circuit_to_data(
        cls, circuit: QuantumCircuit, backend: Backend, convert_wires: bool = True
    ) -> List[List]:
        """Convert the circuit to JSON serializable instructions.

        Helper function that converts a QuantumCircuit into a list of symbolic
        instructions as required by the Json format which is sent to the backend.

        Args:
            circuit: The quantum circuit for which to extract the instructions.
            backend: The backend on which the circuit should be run.
            convert_wires: If True, the circuits are converted to the wiring convention of the backend.

        Returns:
            A list of lists describing the instructions in the circuit. Each sublist
            has three entries the name of the instruction, the wires that the instruction
            applies to and the parameter values of the instruction.
        """

        instructions = []

        config_dict = backend.configuration().to_dict()
        num_species = None
        wire_order = None
        if "num_species" in config_dict:
            num_species = backend.configuration().num_species
            if "wire_order" in config_dict:
                wire_order = WireOrder(backend.configuration().wire_order)
            else:
                wire_order = cls.__wire_order__

        for inst in circuit.data:
            name = inst[0].name
            wires = [circuit.qubits.index(qubit) for qubit in inst[1]]
            if num_species and convert_wires:
                wires = cls.convert_wire_order(
                    wires,
                    convention_from=cls.__wire_order__,
                    convention_to=wire_order,
                    num_species=num_species,
                    num_sites=circuit.num_qubits // num_species,
                    sort=True,
                )
            params = [float(param) for param in inst[0].params]
            instructions.append([name, wires, params])

        return instructions

    @classmethod
    def circuit_to_cold_atom(
        cls,
        circuits: Union[List[QuantumCircuit], QuantumCircuit],
        backend: Backend,
        shots: int = 60,
        convert_wires: bool = True,
    ) -> dict:
        """
        Converts a circuit to a JSon payload to be sent to a given backend.

        Args:
            circuits: The circuits that need to be run.
            backend: The backend on which the circuit should be run.
            shots: The number of shots for each circuit.
            convert_wires: If True, the circuits are converted to the wiring convention of the backend.

        Returns:
            A list of dicts.
        """
        if isinstance(circuits, QuantumCircuit):
            circuits = [circuits]

        # validate the circuits against the backend configuration
        cls.validate_circuits(
            circuits=circuits, backend=backend, shots=shots, convert_wires=convert_wires
        )

        if "wire_order" in backend.configuration().to_dict():
            wire_order = WireOrder(backend.configuration().wire_order)
        else:
            wire_order = cls.__wire_order__

        experiments = {}
        for idx, circuit in enumerate(circuits):
            experiments["experiment_%i" % idx] = {
                "instructions": cls.circuit_to_data(
                    circuit, backend=backend, convert_wires=convert_wires
                ),
                "shots": shots,
                "num_wires": circuit.num_qubits,
                "wire_order": wire_order,
            }

        return experiments

    @classmethod
    def convert_wire_order(
        cls,
        wires: List[int],
        convention_from: WireOrder,
        convention_to: WireOrder,
        num_sites: int,
        num_species: int,
        sort: Optional[bool] = False,
    ) -> List[int]:
        """
        Converts a list of wire indices onto which a gate acts from one convention to another.
        Possible conventions are "sequential", where the first num_sites wires denote the
        first species, the second num_sites wires denote the second species etc., and "interleaved",
        where the first num_species wires denote the first site, the second num_species wires denote
        the second site etc.

        Args:
            wires: Wires onto which a gate acts, e.g. [3, 4, 7, 8].
            convention_from: The convention in which "wires" is given.
            convention_to: The convention into which to convert.
            num_sites: The total number of sites.
            num_species: The number of different atomic species.
            sort: If true, the returned list of indices is sorted in ascending order.

        Raises:
            QiskitColdAtomError: If the convention to and from is not supported.

        Returns:
            A list of wire indices following the convention_to.
        """
        if (convention_to or convention_from) not in WireOrder:
            raise QiskitColdAtomError(
                f"Wire order conversion from {convention_from} to {convention_to}"
                f" is not supported."
            )

        new_wires = None

        if convention_from == convention_to:
            new_wires = wires

        if convention_from == WireOrder.SEQUENTIAL and convention_to == WireOrder.INTERLEAVED:
            new_wires = [idx % num_sites * num_species + idx // num_sites for idx in wires]

        elif convention_from == WireOrder.INTERLEAVED and convention_to == WireOrder.SEQUENTIAL:
            new_wires = [idx % num_species * num_sites + idx // num_species for idx in wires]

        if sort:
            return sorted(new_wires)
        else:
            return new_wires



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Module to build a Fermi-Hubbard problem."""

from abc import ABC, abstractmethod
from typing import List

from qiskit import QuantumCircuit
from qiskit_nature.operators.second_quantization import FermionicOp
from qiskit_cold_atom.fermions.fermion_circuit_solver import FermionicBasis
from qiskit_cold_atom.fermions.fermion_gate_library import FermiHubbard
from qiskit_cold_atom.exceptions import QiskitColdAtomError


class FermionicLattice(ABC):
    """Abstract base fermionic lattice."""

    @property
    @abstractmethod
    def size(self) -> int:
        """The number of lattice sites of the system."""

    @abstractmethod
    def to_fermionic_op(self) -> FermionicOp:
        """Creates the Hamiltonian of the lattice in second quantization.

        Returns:
            The Hamiltonian as a FermionicOp.
        """

    @abstractmethod
    def to_circuit(self, time: float = 1.0) -> QuantumCircuit:
        """
        Wrap the generator of the system in a QuantumCircuit.

        Args:
            time: Duration of the time evolution.

        Returns:
            A quantum circuit which corresponds to the time-evolved Hamiltonian.
        """


class FermiHubbard1D(FermionicLattice):
    """Describes a one-dimensional Fermi-Hubbard model with open boundary conditions."""

    def __init__(
        self,
        num_sites: int,
        particles_up: int,
        particles_down: int,
        hop_strength: float,
        int_strength: float,
        potential: List[float],
    ):
        r"""
        Initialize a one-dimensional fermi-hubbard system. In second quantization this system is
        described by the Hamiltonian

        :math:`H = \sum_{i=1,\sigma}^{L-1} - J_i (f^\dagger_{i,\sigma} f_{i+1,\sigma} +
        f^\dagger_{i+1,\sigma} f_{i,\sigma}) + U \sum_{i=1}^{L} n_{i,\uparrow} n_{i,\downarrow}
        +  \sum_{i=1,\sigma}^{L} \mu_i n_{i,\sigma}`

        Args:
            num_sites: number of lattice sites in the 1D chain.
            particles_up: total number of spin-up particles in the lattice
            particles_down: total number of spin-down particles in the lattice
            hop_strength: strength of hopping between sites
            int_strength: strength of the local interaction
            potential: list of local phases, must be on length num_wires

        Raises:
            QiskitColdAtomError: if the length of the potential does not match the system size.
        """

        # pylint: disable=invalid-name
        self._size = num_sites
        self.particles_up = particles_up
        self.particles_down = particles_down
        self.J = hop_strength
        self.U = int_strength
        self.basis = FermionicBasis(self.size, n_particles=[self.particles_up, self.particles_down])

        if not len(potential) == self.size:
            raise QiskitColdAtomError(
                f"The length of the potentials {len(potential)} must match system size {self.size}"
            )

        self.mu = potential

    @property
    def size(self) -> int:
        """Return the number of sites of the problem."""
        return self._size

    def to_fermionic_op(self) -> FermionicOp:
        """Construct the hamiltonian of the lattice as a FermionicOp.

        Returns:
            A FermionicOp defining the systems Hamiltonian
        """

        operator_labels = []

        # add hopping terms
        for idx in range(self.size - 1):

            right_to_left_up = "I" * idx + "+-" + "I" * (self.size * 2 - idx - 2)
            operator_labels.append((right_to_left_up, -self.J))
            left_to_right_up = "I" * idx + "-+" + "I" * (self.size * 2 - idx - 2)
            operator_labels.append((left_to_right_up, self.J))
            right_to_left_down = "I" * (self.size + idx) + "+-" + "I" * (self.size - idx - 2)
            operator_labels.append((right_to_left_down, -self.J))
            left_to_right_down = "I" * (self.size + idx) + "-+" + "I" * (self.size - idx - 2)
            operator_labels.append((left_to_right_down, self.J))

        # add interaction terms
        for idx in range(self.size):
            opstring = "I" * idx + "N" + "I" * (self.size - 1) + "N" + "I" * (self.size - 1 - idx)
            operator_labels.append((opstring, self.U))

        # add potential terms
        for idx in range(self.size):
            op_up = "I" * idx + "N" + "I" * (2 * self.size - idx - 1)
            operator_labels.append((op_up, self.mu[idx]))
            op_down = "I" * (self.size + idx) + "N" + "I" * (self.size - idx - 1)
            operator_labels.append((op_down, self.mu[idx]))

        return FermionicOp(operator_labels)

    def to_circuit(self, time: float = 1.0) -> QuantumCircuit:
        """
        Wrap the generator of the system in a QuantumCircuit.

        Args:
            time: Duration of the time evolution.

        Returns:
            A quantum circuit which corresponds to the time-evolved Hamiltonian.
        """
        circ = QuantumCircuit(2 * self.size)
        circ.append(
            FermiHubbard(
                num_modes=2 * self.size,
                j=[self.J * time] * (self.size - 1),
                u=self.U * time,
                mu=[mu_i * time for mu_i in self.mu],
            ),
            qargs=range(2 * self.size),
        )

        return circ



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""A solver for time-evolution problems."""

from typing import List

from qiskit_nature.operators.second_quantization import FermionicOp
from qiskit_nature.mappers.second_quantization import (
    JordanWignerMapper,
    BravyiKitaevMapper,
    ParityMapper,
)
from qiskit_nature.converters.second_quantization import QubitConverter

from qiskit import QuantumCircuit, QuantumRegister
from qiskit.opflow.evolutions import PauliTrotterEvolution
from qiskit import execute

from qiskit_cold_atom.applications.fermionic_evolution_problem import (
    FermionicEvolutionProblem,
)
from qiskit_cold_atom.fermions.base_fermion_backend import BaseFermionBackend


class TimeEvolutionSolver:
    """
    Solver class that solves time evolution problem by either analog simulation on fermionic
    hardware or trotterized time evolution on qubit hardware. The computation that this time
    evolution solver will do depends on the type of the backend.
    """

    MAPPER_DICT = {
        "bravyi_kitaev": BravyiKitaevMapper(),
        "jordan_wigner": JordanWignerMapper(),
        "parity": ParityMapper(),
    }

    def __init__(
        self,
        backend,
        map_type: str = None,
        trotter_steps: int = None,
        shots: int = 1000,
    ):
        """
        Initialize a time evolution solver

        Args:
            backend: The backend on which to execute the problem, may be qubit or fermionic.
            map_type: The fermion-to-qubit mapping required if a qubit backend is given
            trotter_steps: The amount of trotter steps to approximate time evolution on
                qubit backends
            shots: number of measurements taken of the constructed circuits
        """

        self.backend = backend
        self.map_type = map_type
        self.trotter_steps = trotter_steps
        self.shots = shots

    def solve(self, problem: FermionicEvolutionProblem) -> List[float]:
        """Solve the problem using the provided backend

        Args:
            problem: The FermionicEvolutionProblem to solve.

        Returns:
            A list of expectation values of the observable of the problem. This list has the
            same length as the list of times for which to compute the time evolution.
        """

        if isinstance(self.backend, BaseFermionBackend):
            qc_load = self.backend.initialize_circuit(problem.initial_state.occupations)
            circuits = problem.circuits(qc_load)

            observable_evs = self.backend.measure_observable_expectation(
                circuits, problem.observable, self.shots
            )

        else:
            # use qubit pipeline
            circuits = self.construct_qubit_circuits(problem)

            # construct observable
            mapper = self.MAPPER_DICT[self.map_type]
            qubit_observable = mapper.map(problem.observable)
            observable_mat = qubit_observable.to_spmatrix()

            observable_evs = [0.0] * len(problem.evolution_times)

            for idx, circuit in enumerate(circuits):

                circuit.measure_all()

                job = execute(circuit, self.backend, shots=self.shots)
                counts = job.result().get_counts().int_outcomes()

                for outcome_ind in counts:
                    prob = counts[outcome_ind] / self.shots

                    observable_evs[idx] += prob * observable_mat.diagonal()[outcome_ind].real

        return observable_evs

    def construct_qubit_circuits(self, problem: FermionicEvolutionProblem) -> List[QuantumCircuit]:
        """Convert the problem to a trotterized qubit circuit using the specified map_type

        Args:
            problem: The fermionic evolution problem specifying the system, evolution-time
                and observable to be measured

        Returns:
            a list of quantum circuits that simulate the time evolution.
            There is one circuit for each evolution time specified in the problem'
        """

        psi_0 = problem.initial_state
        system = problem.system
        hamiltonian = system.to_fermionic_op()

        mapper = self.MAPPER_DICT[self.map_type]

        circuits = []

        # construct circuit of initial state:
        label = ["+" if bit else "I" for bit in psi_0.occupations_flat]
        bitstr_op = FermionicOp("".join(label))
        qubit_op = QubitConverter(mapper).convert(bitstr_op)[0]
        init_circ = QuantumCircuit(QuantumRegister(qubit_op.num_qubits, "q"))

        for i, pauli_label in enumerate(qubit_op.primitive.paulis[0].to_label()[::-1]):
            if pauli_label == "X":
                init_circ.x(i)
            elif pauli_label == "Y":
                init_circ.y(i)
            elif pauli_label == "Z":
                init_circ.z(i)

        for time in problem.evolution_times:

            # time-step of zero will cause PauliTrotterEvolution to fail
            if time == 0.0:
                time += 1e-10

            # map fermionic hamiltonian to qubits
            qubit_hamiltonian = mapper.map(hamiltonian * time)
            # get time evolution operator by exponentiating
            exp_op = qubit_hamiltonian.exp_i()
            # perform trotterization

            evolved_op = PauliTrotterEvolution(reps=self.trotter_steps).convert(exp_op)

            trotter_circ = evolved_op.to_circuit_op().to_circuit()

            circuits.append(init_circ.compose(trotter_circ))

        return circuits



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""A transpiler pass to simplify single-qubit gates."""

from qiskit.transpiler.basepasses import TransformationPass
from qiskit.transpiler.exceptions import TranspilerError

from qiskit_cold_atom.spins.spins_gate_library import (
    RLXGate,
    RLYGate,
    RLZGate,
    RLZ2Gate,
)


class Optimize1SpinGates(TransformationPass):
    """Simplify single-spin spin gates.

    The simplification of these gates is done by adding the angle of identical and
    consecutive gates together. For example, the circuit

    .. parsed-literal::

                ┌──────────┐┌──────────┐┌──────────┐┌──────────┐┌──────────┐┌──────────┐
        spin_0: ┤ Rlx(π/2) ├┤ Rlx(π/2) ├┤ Rlx(π/2) ├┤ Rly(π/2) ├┤ Rly(π/2) ├┤ Rlx(π/2) ├
                └──────────┘└──────────┘└──────────┘└──────────┘└──────────┘└──────────┘

    will be transpiled to

    .. parsed-literal::

                ┌───────────┐┌────────┐┌──────────┐
        spin_0: ┤ Rlx(3π/2) ├┤ Rly(π) ├┤ Rlx(π/2) ├
                └───────────┘└────────┘└──────────┘
    """

    def __init__(self):
        super().__init__()
        self._gate_names = ["rlx", "rly", "rlz", "rlz2"]

    def run(self, dag):
        """Run the Optimize1qSpinGates pass on `dag`.

        Args:
            dag (DAGCircuit): the DAG to be optimized.

        Returns:
            DAGCircuit: the optimized DAG.

        Raises:
            TranspilerError: if the considered gate is not supported.
        """

        for gate_name in self._gate_names:
            runs = dag.collect_runs(gate_name)

            for run in runs:
                if len(run) == 1:
                    continue

                # Sum the angles of the same single-qubit gates.
                total_angle = 0.0
                for dag_node in run:
                    total_angle += dag_node.op.params[0]

                if gate_name == "rlx":
                    new_op = RLXGate(total_angle)
                elif gate_name == "rly":
                    new_op = RLYGate(total_angle)
                elif gate_name == "rlz":
                    new_op = RLZGate(total_angle)
                elif gate_name == "rlz2":
                    new_op = RLZ2Gate(total_angle)
                else:
                    raise TranspilerError(f"Could not use the basis {self._gate_names}.")

                dag.substitute_node(run[0], new_op, inplace=True)

                # Delete the remaining nodes
                for current_node in run[1:]:
                    dag.remove_op_node(current_node)

        return dag



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Module for cold-atom spin backends."""

from abc import ABC

from qiskit.providers import BackendV1 as Backend
from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister
from qiskit_cold_atom import QiskitColdAtomError


class BaseSpinBackend(Backend, ABC):
    """Abstract base class for atomic mixture backends."""

    def get_empty_circuit(self) -> QuantumCircuit:
        """
        Convenience function to set up an empty circuit with the right QuantumRegisters.
        For each atomic species specified in the config file, a quantum register is added to the circuit.

        Returns:
            qc: An empty quantum circuit ready to use in spin-based cold-atom setups.

        Raises:
            QiskitColdAtomError:
                - If backend has no config file.
                - If number of wires of the backend config is not a multiple of the atomic species.
        """
        config = self.configuration().to_dict()

        try:
            num_wires = config["n_qubits"]
            num_species = len(config["atomic_species"])

        except NameError as name_error:
            raise QiskitColdAtomError(
                "backend needs to be initialized with config file first"
            ) from name_error

        if not (isinstance(num_wires, int) and num_wires % num_species == 0):
            raise QiskitColdAtomError(
                "num_wires {num_wires} must be multiple of num_species {num_species}"
            )

        qregs = [
            QuantumRegister(num_wires / num_species, species)
            for species in config["atomic_species"]
        ]

        class_reg = ClassicalRegister(num_wires, "c{}".format(num_wires))
        empty_circuit = QuantumCircuit(*qregs, class_reg)

        return empty_circuit

    def draw(self, qc: QuantumCircuit, **draw_options):
        """Modified circuit drawer to better display atomic mixture quantum circuits.

        For now this method is just an alias to `QuantumCircuit.draw()` but in the future this method
        may be modified and tailored to spin quantum circuits.

        Args:
            qc: The quantum circuit to draw.
            draw_options: Key word arguments for the drawing of circuits.
        """
        qc.draw(**draw_options)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

r"""
.. image:: ../images/qiskit_cold_atom_logo_with_text.svg
   :alt: Missing Logo

==================================================
Qiskit Cold Atom module (:mod:`qiskit_cold_atom`)
==================================================

.. currentmodule:: qiskit_cold_atom

The Qiskit Cold Atom module provides functionality to describe quantum systems of trapped cold atoms
in a gate- and circuit-based framework.

Traditionally, each wire in a quantum circuit represents one qubit as the fundamental unit of information
processing. Here, we extend this concept and allow wires to represent individual internal states of
trapped cold atoms. This currently covers two settings, one for fermionic modes and one for spin
modes.

In a fermionic setting, each wire of a quantum circuit represents an abstract fermionic mode in second
quantization which can either be occupied (1) or empty (0). Such systems are realized experimentally by
individual fermionic atoms trapped in arrays of optical tweezers. Circuit instructions and backends
to interact with and simulate such circuits are given by the :mod:`qiskit_cold_atom.fermions` module.

In a spin setting, each wire of a quantum circuit represents a quantum mechanical spin of a given length
:math:`S`. Upon measurement, each spin is measured in one of its :math:`2S+1` internal basis states
labelled :math:`0` to :math:`2S`, thus it can be thought of as a qudit with dimension :math:`d = 2S+1`.
This setting describes the collective spin of bosonic atoms trapped in a Bose-Einstein-condensate.
Circuit instructions and backends to interact with and simulate such circuits are provided by the
:mod:`qiskit_cold_atom.spins` module.

The quantum circuits that these systems can implement thus utilize a fundamentally different form of
quantum information processing compared to qubits. Therefore, the typical qubit gates can not be applied
to these circuits. Instead, the fermions and spin modules define their own gate sets which are defined
by their second-quantized Hamiltonians that generate the unitary gate. Note that loading the
:mod:`qiskit_cold_atom.fermions` or :mod:`qiskit_cold_atom.spins` module will decorate the
:class:`QuantumCircuit` class in Qiskit by adding methods to call pre-defined fermionic and spin gates,
respectively.

To enable the control of real quantum hardware, the :mod:`qiskit_cold_atom.providers`
module contains a provider which enables access to cold atomic device backends.

The top-level classes and submodules of qiskit_cold_atom are:

.. autosummary::
   :toctree: ../stubs/
   :nosignatures:

   QiskitColdAtomError

Submodules
==========

.. autosummary::
   :toctree:

   applications
   fermions
   providers
   spins

"""
from functools import wraps
from qiskit import QuantumCircuit
from qiskit_cold_atom.exceptions import QiskitColdAtomError


def add_gate(func):
    """Decorator to add a gate method to the QuantumCircuit class"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    setattr(QuantumCircuit, func.__name__, wrapper)

    return func


__version__ = "0.1.0"

__all__ = ["__version__", "QiskitColdAtomError", "add_gate"]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Module for cold-atom fermion backends."""

from abc import ABC
from typing import Union, List, Optional
import numpy as np

from qiskit.providers import BackendV1 as Backend
from qiskit import QuantumCircuit
from qiskit_nature.operators.second_quantization import FermionicOp

from qiskit_cold_atom.fermions.fermion_gate_library import LoadFermions
from qiskit_cold_atom.fermions.fermion_circuit_solver import FermionCircuitSolver
from qiskit_cold_atom.fermions.fermionic_state import FermionicState
from qiskit_cold_atom.exceptions import QiskitColdAtomError


class BaseFermionBackend(Backend, ABC):
    """Abstract base class for fermionic tweezer backends."""

    def initialize_circuit(self, occupations: Union[List[int], List[List[int]]]):
        """
        Initialize a fermionic quantum circuit with the given occupations.

        Args:
            occupations: List of occupation numbers. When ``List[int]`` is given, the occupations
                correspond to the number of indistinguishable fermionic particles in each mode,
                e.g. ``[0, 1, 1, 0]`` implies that sites one and two are occupied by a fermion.
                When ``List[List[int]]`` is given, the occupations describe the number of particles in
                fermionic modes with different (distinguishable) species of fermions. Each
                inner list gives the occupations of one fermionic species.

        Returns:
            circuit: Qiskit QuantumCircuit with a quantum register for each fermionic species
                     initialized with the ``load`` instructions corresponding to the given occupations

        Raises:
            QiskitColdAtomError: If occupations do not match the backend
        """
        try:
            backend_size = self.configuration().to_dict()["n_qubits"]
        except NameError as name_error:
            raise QiskitColdAtomError(
                f"Number of tweezers not specified for {self.name()}"
            ) from name_error

        initial_state = FermionicState(occupations)

        n_wires = initial_state.sites * initial_state.num_species

        if n_wires > backend_size:
            raise QiskitColdAtomError(
                f"{self.name()} supports up to {backend_size} sites, {n_wires} were given"
            )

        # if num_species is specified by the backend, the wires describe different atomic species
        # and the circuit must exactly match the expected wire count of the backend.
        if "num_species" in self.configuration().to_dict().keys():
            num_species = self.configuration().num_species
            if num_species > 1 and n_wires < self.configuration().num_qubits:
                raise QiskitColdAtomError(
                    f"{self.name()} requires circuits with exactly "
                    f"{self.configuration().num_qubits} wires, but an initial occupation of size "
                    f"{n_wires} was given."
                )

        from qiskit.circuit import QuantumRegister

        if initial_state.num_species > 1:
            registers = []
            for i in range(initial_state.num_species):
                registers.append(QuantumRegister(initial_state.sites, f"spin_{i}"))
            circuit = QuantumCircuit(*registers)

        else:
            circuit = QuantumCircuit(QuantumRegister(initial_state.sites, "fer_mode"))

        for i, occupation_list in enumerate(initial_state.occupations):
            for j, occ in enumerate(occupation_list):
                if occ:
                    circuit.append(LoadFermions(), qargs=[i * initial_state.sites + j])

        return circuit

    def measure_observable_expectation(
        self,
        circuits: Union[QuantumCircuit, List[QuantumCircuit]],
        observable: FermionicOp,
        shots: int,
        seed: Optional[int] = None,
        num_species: int = 1,
        get_variance: bool = False,
    ):
        """Measure the expectation value of an observable in a state prepared by a given quantum circuit
        that uses fermionic gates. Measurements are added to the entire register if they are not yet
        applied in the circuit.

        Args:
            circuits: QuantumCircuit applying gates with fermionic generators
            observable: A FermionicOp describing an observable of which the expectation value is sampled
            shots: Number of measurement shots taken in case the circuit has measure instructions
            seed: seed for the random number generator of the measurement simulation
            num_species: number of different fermionic species described by the circuits
            get_variance: If True, also returns an estimate of the variance of the observable

        Raises:
            QiskitColdAtomError: if the observable is non-diagonal

        Returns:
            observable_ev: List of the measured expectation values of the observables in given circuits
            variance: List of the estimated variances of of the observables (if get_variance is True)
        """

        if isinstance(circuits, QuantumCircuit):
            circuits = [circuits]

        observable_evs = [0] * len(circuits)
        observable_vars = [0] * len(circuits)

        for idx, circuit in enumerate(circuits):

            # check whether the observable is diagonal in the computational basis.
            solver = FermionCircuitSolver(num_species=2)
            solver.preprocess_circuit(circuit)
            observable_mat = solver.operator_to_mat(observable)

            if list(observable_mat.nonzero()[0]) != list(observable_mat.nonzero()[1]):
                raise QiskitColdAtomError(
                    "Measuring general observables that are non-diagonal in the "
                    "computational basis is not yet implemented for "
                    "fermionic backends. This requires non-trivial basis "
                    "transformations that are in general difficult to find and "
                    "depend on the backend's native gate set."
                )

            circuit.remove_final_measurements()
            circuit.measure_all()

            # pylint: disable=unexpected-keyword-arg
            job = self.run(circuit, shots=shots, seed=seed, num_species=num_species)
            counts = job.result().get_counts()

            for bitstring in counts:
                # Extract the index of the measured count-bitstring in the fermionic basis.
                # In contrast to qubits, this is not trivial and requires an additional step.
                ind = solver.basis.get_index_of_measurement(bitstring)

                # contribution to the operator estimate of this outcome
                p = counts[bitstring] / shots
                observable_evs[idx] += p * observable_mat[ind, ind].real

                if get_variance:
                    # contribution to the variance of the operator
                    observable_vars[idx] += (
                        np.sqrt(p * (1 - p) / shots) * observable_mat[ind, ind]
                    ) ** 2

        if get_variance:
            return observable_evs, observable_vars
        else:
            return observable_evs

    def draw(self, qc: QuantumCircuit, **draw_options):
        """Modified circuit drawer to better display atomic mixture quantum  circuits.

        Note that in the future this method may be modified and tailored to fermionic quantum circuits.

        Args:
            qc: The quantum circuit to draw.
            draw_options: Key word arguments for the drawing of circuits.
        """
        qc.draw(**draw_options)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

""" Spin gate tests."""

import numpy as np
from scipy.linalg import expm

from qiskit.test import QiskitTestCase
from qiskit import QuantumCircuit
from qiskit_nature.operators.second_quantization import SpinOp

from qiskit_cold_atom.spins.spin_circuit_solver import SpinCircuitSolver
from qiskit_cold_atom.spins import SpinSimulator
from qiskit_cold_atom.spins.spins_gate_library import (
    RLXGate,
    RLYGate,
    RLZGate,
    RLZ2Gate,
    RLXLYGate,
    RydbergBlockade,
    RydbergFull,
)


class TestSpinGates(QiskitTestCase):
    """Tests for the spin hardware gates"""

    def setUp(self):
        super().setUp()
        self.backend = SpinSimulator()
        self.spin = 3 / 2
        self.solver = SpinCircuitSolver(spin=self.spin)

    def test_lx_gate(self):
        """check matrix form of the lx gate"""
        omega = np.pi / 2
        circ = QuantumCircuit(1)
        circ.append(RLXGate(omega), qargs=[0])
        # add gate to circuit via the @add_gate-decorated method
        circ_decorated = QuantumCircuit(1)
        circ_decorated.rlx(omega, 0)

        for circuit in [circ, circ_decorated]:
            unitary = self.backend.run(circuit, spin=self.spin).result().get_unitary()
            self.assertTrue(
                np.allclose(
                    unitary,
                    expm(
                        -1j
                        * omega
                        * np.array(
                            [
                                [0.0, np.sqrt(3) / 2, 0, 0],
                                [np.sqrt(3) / 2, 0, 1, 0],
                                [0, 1, 0, np.sqrt(3) / 2],
                                [0, 0, np.sqrt(3) / 2, 0],
                            ]
                        )
                    ),
                )
            )

    def test_lxly_gate(self):
        """check matrix form of the lxly gate"""
        omega = np.pi
        circ = QuantumCircuit(2)
        circ.append(RLXLYGate(omega), qargs=[0, 1])

        # add gate to circuit via the @add_gate-decorated method
        circ_decorated = QuantumCircuit(2)
        circ_decorated.rlxly(omega, [0, 1])

        for circuit in [circ, circ_decorated]:
            unitary = self.backend.run(circuit, spin=1 / 2).result().get_unitary()
            self.assertTrue(
                np.allclose(
                    unitary,
                    [
                        [1, 0, 0, 0],
                        [0, 0, -1j, 0],
                        [0, -1j, 0, 0],
                        [0, 0, 0, 1],
                    ],
                )
            )

    def test_ly_gate(self):
        """check matrix form of the ly gate"""
        omega = np.pi / 2
        circ = QuantumCircuit(1)
        circ.append(RLYGate(omega), qargs=[0])
        # add gate to circuit via the @add_gate-decorated method
        circ_decorated = QuantumCircuit(1)
        circ_decorated.rly(omega, 0)

        for circuit in [circ, circ_decorated]:
            unitary = self.backend.run(circuit, spin=self.spin).result().get_unitary()
            self.assertTrue(
                np.allclose(
                    unitary,
                    expm(
                        -1j
                        * omega
                        * np.array(
                            [
                                [0.0, -1j * np.sqrt(3) / 2, 0, 0],
                                [1j * np.sqrt(3) / 2, 0, -1j, 0],
                                [0, 1j, 0, -1j * np.sqrt(3) / 2],
                                [0, 0, 1j * np.sqrt(3) / 2, 0],
                            ]
                        )
                    ),
                )
            )

    def test_lz_gate(self):
        """check matrix form of the lz gate"""
        delta = np.pi / 2
        circ = QuantumCircuit(1)
        circ.append(RLZGate(delta), qargs=[0])
        # add gate to circuit via the @add_gate-decorated method
        circ_decorated = QuantumCircuit(1)
        circ_decorated.rlz(delta, 0)

        for circuit in [circ, circ_decorated]:
            unitary = self.backend.run(circuit, spin=self.spin).result().get_unitary()
            self.assertTrue(
                np.allclose(
                    unitary,
                    expm(
                        -1j
                        * delta
                        * np.array(
                            [
                                [3 / 2, 0, 0, 0],
                                [0, 1 / 2, 0, 0],
                                [0, 0, -1 / 2, 0],
                                [0, 0, 0, -3 / 2],
                            ]
                        )
                    ),
                )
            )

    def test_lz2_gate(self):
        """check matrix form of the lz2 gate"""
        chi = np.pi / 2
        circ = QuantumCircuit(1)
        circ.append(RLZ2Gate(chi), qargs=[0])
        # add gate to circuit via the @add_gate-decorated method
        circ_decorated = QuantumCircuit(1)
        circ_decorated.rlz2(chi, 0)

        for circuit in [circ, circ_decorated]:
            unitary = self.backend.run(circuit, spin=self.spin).result().get_unitary()
            self.assertTrue(
                np.allclose(
                    unitary,
                    expm(
                        -1j
                        * chi
                        * np.array(
                            [
                                [3 / 2, 0, 0, 0],
                                [0, 1 / 2, 0, 0],
                                [0, 0, -1 / 2, 0],
                                [0, 0, 0, -3 / 2],
                            ]
                        )
                        ** 2
                    ),
                )
            )

    def test_rydberg_block_gate(self):
        """check matrix form of the rydberg blockade gate on two qubits"""
        chi = 2 * np.pi

        # the expected unitary is the following
        expected = np.exp(1j * np.pi / 2) * expm(
            -1j
            * chi
            * np.array(
                [
                    [0, 0, 0, 0],
                    [0, 0, 0, 0],
                    [0, 0, 0, 0],
                    [0, 0, 0, 1],
                ]
            )
        )
        # first test the RydbergBlockade only
        with self.subTest("test generation of operator"):
            from qiskit.quantum_info import Operator

            self.assertTrue(np.allclose(Operator(RydbergBlockade(2, phi=chi)).data, expected))

        # add gate to circuit via the @add_gate-decorated method
        circ_decorated = QuantumCircuit(2)
        circ_decorated.rydberg_block(chi, [0, 1])
        unitary = self.backend.run(circ_decorated).result().get_unitary()
        self.assertTrue(np.allclose(unitary, expected))

    def test_rydberg_full_gate(self):
        """check matrix form of the full rydberg Hamiltonian on two qubits"""
        chi = 2 * np.pi
        circ = QuantumCircuit(2)
        circ.append(RydbergFull(2, omega=0, delta=0, phi=chi), qargs=[0, 1])

        # add gate to circuit via the @add_gate-decorated method
        circ_decorated = QuantumCircuit(2)
        circ_decorated.rydberg_full(omega=0, delta=0, phi=chi, modes=range(2))
        for circuit in [circ, circ_decorated]:
            unitary = self.backend.run(circuit).result().get_unitary()
            self.assertTrue(
                np.allclose(
                    unitary,
                    np.exp(1j * np.pi / 2)
                    * expm(
                        -1j
                        * chi
                        * np.array(
                            [
                                [0, 0, 0, 0],
                                [0, 0, 0, 0],
                                [0, 0, 0, 0],
                                [0, 0, 0, 1],
                            ]
                        )
                    ),
                )
            )

        # also test the sigma gates
        circ = QuantumCircuit(2)
        circ.append(RydbergFull(2, omega=np.pi, delta=0, phi=0), qargs=[0, 1])
        # add gate to circuit via the @add_gate-decorated method
        circ_decorated = QuantumCircuit(2)
        circ_decorated.rydberg_full(omega=np.pi, delta=0, phi=0, modes=range(2))
        for circuit in [circ, circ_decorated]:
            unitary = self.backend.run(circuit).result().get_unitary()
            self.assertTrue(
                np.allclose(
                    unitary,
                    expm(
                        -1j
                        * np.pi
                        / 2
                        * np.array(
                            [
                                [0, 1, 1, 0],
                                [1, 0, 0, 1],
                                [1, 0, 0, 1],
                                [0, 1, 1, 0],
                            ]
                        )
                    ),
                )
            )

        # also test the sigmaz gates
        circ = QuantumCircuit(2)
        circ.append(RydbergFull(2, omega=0, delta=np.pi, phi=0), qargs=[0, 1])
        # add gate to circuit via the @add_gate-decorated method
        circ_decorated = QuantumCircuit(2)
        circ_decorated.rydberg_full(omega=0, delta=np.pi, phi=0, modes=range(2))
        for circuit in [circ, circ_decorated]:
            unitary = self.backend.run(circuit).result().get_unitary()
            self.assertTrue(
                np.allclose(
                    unitary,
                    expm(
                        -1j
                        * np.pi
                        / 2
                        * np.array(
                            [
                                [-2, 0, 0, 0],
                                [0, 0, 0, 0],
                                [0, 0, 0, 0],
                                [0, 0, 0, 2],
                            ]
                        )
                    ),
                )
            )

    def test_spin_gate(self):
        """test the functionality of the base class for fermionic gates"""
        test_gates = [RLXGate(0.8), RLYGate(2.4), RLZGate(5.6), RLZ2Gate(1.3)]
        with self.subTest("test to_matrix and power"):
            for gate in test_gates:
                exp_matrix = gate.to_matrix() @ gate.to_matrix()
                exp_gate = gate.power(2)
                self.assertTrue(np.allclose(exp_matrix, exp_gate.to_matrix()))

        with self.subTest("test generation of operator"):
            from qiskit.quantum_info import Operator

            for gate in test_gates:
                self.assertTrue(isinstance(Operator(gate), Operator))

    def test_identity_gates(self):
        """test that gates with parameters equal to zero still have a well-defined generator."""
        test_gates = [RLXGate(0.0), RLYGate(0.0), RLZGate(0.0), RLZ2Gate(0.0)]

        for gate in test_gates:
            self.assertIsInstance(gate.generator, SpinOp)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Classes for remote cold atom backends."""

import json
from typing import List, Dict, Union
import requests

from qiskit.providers.models import BackendConfiguration
from qiskit.providers import Options
from qiskit import QuantumCircuit
from qiskit.providers import ProviderV1 as Provider
from qiskit.providers import BackendV1 as Backend
from qiskit.providers import JobStatus

from qiskit_cold_atom.spins.base_spin_backend import BaseSpinBackend
from qiskit_cold_atom.fermions.base_fermion_backend import BaseFermionBackend
from qiskit_cold_atom.circuit_tools import CircuitTools
from qiskit_cold_atom.providers.cold_atom_job import ColdAtomJob
from qiskit_cold_atom.exceptions import QiskitColdAtomError


class RemoteBackend(Backend):
    """Remote cold atom backend."""

    def __init__(self, provider: Provider, url: str):
        """
        Initialize the backend by querying the server for the backend configuration dictionary.

        Args:
            provider: The provider which need to have the correct credentials attributes in place
            url: The url of the backend server

        Raises:
            QiskitColdAtomError: If the connection to the backend server can not be established.

        """

        self.url = url
        self.username = provider.credentials["username"]
        self.token = provider.credentials["token"]
        # Get the config file from the remote server
        try:
            r = requests.get(
                self.url + "/get_config",
                params={
                    "username": self.username,
                    "token": self.token,
                },
            )
        except requests.exceptions.ConnectionError as err:
            raise QiskitColdAtomError(
                "connection to the backend server can not be established."
            ) from err

        super().__init__(configuration=BackendConfiguration.from_dict(r.json()), provider=provider)

    @classmethod
    def _default_options(cls) -> Options:
        """Return the default options.

        Returns:
            qiskit.providers.Options: A options object with default values set
        """
        return Options(shots=1)

    @property
    def credentials(self) -> Dict[str, Union[str, List[str]]]:
        """Returns: the access credentials used."""
        return self.provider().credentials

    # pylint: disable=arguments-differ, unused-argument
    def run(
        self,
        circuit: Union[QuantumCircuit, List[QuantumCircuit]],
        shots: int = 1,
        convert_wires: bool = True,
        **run_kwargs,
    ) -> ColdAtomJob:
        """
        Run a quantum circuit or list of quantum circuits.

        Args:
            circuit: The quantum circuits to be executed on the device backend
            shots: The number of measurement shots to be measured for each given circuit
            convert_wires: If True (the default), the circuits are converted to the wiring convention
            of the backend.
            run_kwargs: Additional keyword arguments that might be passed down when calling
            qiskit.execute() which will have no effect on this backend.

        Raises:
            QiskitColdAtomError: If the response from the backend does not have a job_id.

        Returns:
            A Job object through the backend can be queried for status, result etc.
        """

        job_payload = CircuitTools.circuit_to_cold_atom(circuit, self, shots=shots)

        res = requests.post(
            self.url + "/post_job",
            json={
                "job": json.dumps(job_payload),
                "username": self.username,
                "token": self.token,
            },
        )

        res.raise_for_status()
        response = res.json()

        if "job_id" not in response:
            raise QiskitColdAtomError("The response has no job_id.")

        return ColdAtomJob(self, response["job_id"])

    def retrieve_job(self, job_id: str) -> ColdAtomJob:
        """Return a single job submitted to this backend.

        Args:
            job_id: The ID of the job to retrieve.

        Returns:
            The job with the given ID.

        Raises:
            QiskitColdAtomError: If the job retrieval failed.
        """

        retrieved_job = ColdAtomJob(backend=self, job_id=job_id)

        try:
            job_status = retrieved_job.status()
        except requests.exceptions.RequestException as request_error:
            raise QiskitColdAtomError(
                "connection to the remote backend could not be established"
            ) from request_error

        if job_status == JobStatus.ERROR:
            raise QiskitColdAtomError(f"Job with id {job_id} could not be retrieved")

        return retrieved_job


class RemoteSpinBackend(RemoteBackend, BaseSpinBackend):
    """Remote backend which runs spin circuits."""


class RemoteFermionBackend(RemoteBackend, BaseFermionBackend):
    """Remote backend which runs fermionic circuits."""



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Fermionic circuit solver tests."""

import numpy as np

from qiskit import QuantumCircuit
from qiskit.test import QiskitTestCase
from qiskit_nature.operators.second_quantization import FermionicOp, SpinOp

from qiskit_cold_atom.fermions.fermion_circuit_solver import (
    FermionCircuitSolver,
    FermionicBasis,
)
from qiskit_cold_atom.fermions.fermion_gate_library import Hop
from qiskit_cold_atom.exceptions import QiskitColdAtomError

# Black import needed to decorate the quantum circuit with spin gates.
import qiskit_cold_atom.spins  # pylint: disable=unused-import


class TestFermionCircuitSolver(QiskitTestCase):
    """class to test the FermionCircuitSolver class."""

    def setUp(self):
        super().setUp()
        # Setup two solvers with different number of fermionic species
        self.solver1 = FermionCircuitSolver()
        self.solver2 = FermionCircuitSolver(num_species=2)

    def test_basis_setter(self):
        """test max. dimension of the basis"""
        self.solver1.max_dimension = 500
        with self.assertRaises(QiskitColdAtomError):
            self.solver1.basis = FermionicBasis(sites=12, n_particles=6)

    def test_preprocess_circuit(self):
        """test the preprocessing of the circuit"""
        circ = QuantumCircuit(4, 4)
        circ.fload([0, 2])
        circ.fhop([0.5], [0, 1, 2, 3])
        with self.subTest("spin conserving circuit"):
            self.solver2.preprocess_circuit(circ)
            self.assertEqual(self.solver2._dim, 4)
        with self.subTest("non spin conserving circuit"):
            self.solver1.preprocess_circuit(circ)
            self.assertEqual(self.solver1._dim, 6)

    def test_get_initial_state(self):
        """test initialization of the state for the simulation"""
        circ = QuantumCircuit(4, 4)
        circ.fload([0, 3])
        self.solver2.preprocess_circuit(circ)
        init_state = self.solver2.get_initial_state(circ)
        target = np.array([0, 0, 1, 0])
        self.assertTrue(np.alltrue(init_state.toarray().T == target))

    def test_embed_operator(self):
        """test embedding of an operator"""
        fer_op = FermionicOp("+-")
        spin_op = SpinOp("+-")
        num_wires = 4
        qargs = [1, 3]
        qargs_wrong = [0, 1, 3]

        with self.subTest("check operator type"):
            with self.assertRaises(QiskitColdAtomError):
                self.solver1._embed_operator(spin_op, num_wires, qargs)

        with self.subTest("check operator wiring"):
            with self.assertRaises(QiskitColdAtomError):
                self.solver1._embed_operator(fer_op, num_wires, qargs_wrong)

        with self.subTest("operator embedding"):
            embedded_op = self.solver1._embed_operator(fer_op, num_wires, qargs)
            target_op = FermionicOp("+_1 -_3", register_length=4)
            self.assertTrue(
                set(embedded_op.simplify().to_list(display_format="dense"))
                == set(target_op.simplify().to_list(display_format="dense"))
            )

    def test_conservation_checks(self):
        """test the checks for conservation of spin-species."""
        with self.subTest("check operator type"):
            circ = QuantumCircuit(4, 4)
            circ.fhop([0.5], [0, 1, 2, 3])
            circ.rlx(0.5, 0)  # apply gate with a SpinOp generator
            with self.assertRaises(QiskitColdAtomError):
                self.solver1._check_conservations(circ)

        with self.subTest("check compatibility with number of species"):
            circ = QuantumCircuit(5, 5)
            circ.fhop([0.5], [0, 1, 2, 3])
            self.assertTrue(self.solver1._check_conservations(circ) == (True, True))
            with self.assertRaises(QiskitColdAtomError):
                self.solver2._check_conservations(circ)

        with self.subTest("spin conserved"):
            circ = QuantumCircuit(4, 4)
            circ.fload([0, 3])
            circ.fhop([0.5], [0, 1, 2, 3])
            self.assertTrue(self.solver2._check_conservations(circ) == (True, True))

        with self.subTest("spin not conserved"):
            circ = QuantumCircuit(4, 4)
            circ.fload([0, 3])
            circ.fhop([0.5], [0, 1, 2, 3])
            circ.frx(0.3, [0, 2])  # non spin-conserving gate
            self.assertTrue(self.solver2._check_conservations(circ) == (True, False))

    def test_operator_to_mat(self):
        """test matrix representation of fermionic gates"""

        with self.subTest("check operator type"):
            spin_op = SpinOp("+-")
            with self.assertRaises(QiskitColdAtomError):
                self.solver1.operator_to_mat(spin_op)

        circ = QuantumCircuit(4, 4)
        circ.fload([0, 3])
        circ.fhop([0.5], [0, 1, 2, 3])

        with self.subTest("check dimensionality of operator"):
            self.solver2.preprocess_circuit(circ)
            fer_op_wrong = FermionicOp("+-I")
            fer_op_correct = FermionicOp("+-II", register_length=4)
            with self.assertRaises(QiskitColdAtomError):
                self.solver2.operator_to_mat(fer_op_wrong)
            self.solver2.operator_to_mat(fer_op_correct)

        with self.subTest("test matrix representation"):
            self.solver2.preprocess_circuit(circ)
            target = np.array(
                [
                    [0.0, -0.5, -0.5, 0.0],
                    [-0.5, 0.0, 0.0, -0.5],
                    [-0.5, 0.0, 0.0, -0.5],
                    [0.0, -0.5, -0.5, 0.0],
                ]
            )
            test_op = self.solver2.operator_to_mat(Hop(num_modes=4, j=[0.5]).generator)
            self.assertTrue(np.alltrue(test_op.toarray() == target))

    def test_draw_shots(self):
        """test drawing of the shots from a measurement distribution"""
        circ = QuantumCircuit(4, 4)
        circ.fload([0, 3])
        circ.fhop([0.5], [0, 1, 2, 3])
        self.solver2.preprocess_circuit(circ)

        with self.subTest("check missing shot number"):
            # error because the number of shots is not specified
            with self.assertRaises(QiskitColdAtomError):
                self.solver2.draw_shots(np.ones(4) / 4)

        self.solver2.shots = 5

        with self.subTest("check match of dimensions"):
            # error because there is a mismatch in the dimension
            with self.assertRaises(QiskitColdAtomError):
                self.solver2.draw_shots(np.ones(3) / 3)

        with self.subTest("formatting of measurement outcomes"):

            self.solver2.seed = 40
            outcomes = self.solver2.draw_shots(np.ones(4) / 4)
            self.assertEqual(outcomes, ["0110", "0101", "1010", "0110", "0110"])

    def test_to_operators(self):
        """test the to_operators method inherited form BaseCircuitSolver"""

        test_circ = QuantumCircuit(4, 4)
        test_circ.fload([0, 3])
        test_circ.fhop([0.5], [0, 1, 2, 3])
        test_circ.fint(1.0, [0, 1, 2, 3])
        test_circ.measure_all()

        with self.subTest("test ignore barriers"):
            self.solver1.ignore_barriers = False
            with self.assertRaises(NotImplementedError):
                self.solver1.to_operators(test_circ)
            self.solver1.ignore_barriers = True

        with self.subTest("check for gate generators"):
            qubit_circ = QuantumCircuit(1)
            qubit_circ.h(0)
            with self.assertRaises(QiskitColdAtomError):
                self.solver1.to_operators(qubit_circ)

        with self.subTest("gate after previous measurement instruction"):
            meas_circ = QuantumCircuit(4, 4)
            meas_circ.measure_all()
            meas_circ.fhop([0.5], [0, 1, 2, 3])
            with self.assertRaises(QiskitColdAtomError):
                self.solver1.to_operators(meas_circ)

        with self.subTest("check returned operators"):
            operators = self.solver1.to_operators(test_circ)
            target = [
                FermionicOp(
                    [
                        ("+-II", -0.5),
                        ("-+II", 0.5),
                        ("II+-", -0.5),
                        ("II-+", 0.5),
                    ]
                ),
                FermionicOp([("NINI", 1), ("ININ", 1)]),
            ]
            for i, op in enumerate(operators):
                self.assertEqual(
                    set(op.simplify().to_list(display_format="dense")),
                    set(target[i].simplify().to_list(display_format="dense")),
                )

    def test_call_method(self):
        """test the call method inherited form BaseCircuitSolver that simulates a circuit"""

        test_circ = QuantumCircuit(4)
        test_circ.fload([0, 3])
        test_circ.fhop([np.pi / 4], [0, 1, 2, 3])
        test_circ.fint(np.pi, [0, 1, 2, 3])

        with self.subTest("running the circuit"):
            self.solver2.shots = 5
            self.solver2.seed = 40
            simulation = self.solver2(test_circ)

            self.assertEqual(simulation["memory"], ["0110", "0101", "1010", "0110", "0110"])
            self.assertEqual(simulation["counts"], {"1010": 1, "0110": 3, "0101": 1})
            self.assertTrue(
                np.allclose(simulation["statevector"], np.array([-0.5j, -0.5, 0.5, -0.5j]))
            )
            self.assertTrue(
                np.allclose(
                    simulation["unitary"],
                    np.array(
                        [
                            [-0.5, -0.5j, -0.5j, 0.5],
                            [0.5j, 0.5, -0.5, 0.5j],
                            [0.5j, -0.5, 0.5, 0.5j],
                            [0.5, -0.5j, -0.5j, -0.5],
                        ]
                    ),
                )
            )

        with self.subTest("check for maximum dimension"):
            self.solver2.max_dimension = 3
            with self.assertRaises(QiskitColdAtomError):
                simulation = self.solver2(test_circ)
            self.solver2.max_dimension = 100

        with self.subTest("check if shots are specified"):
            self.solver2.shots = None
            simulation = self.solver2(test_circ)
            self.assertEqual(simulation["memory"], [])
            self.assertEqual(simulation["counts"], {})



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Module to simulate fermionic circuits."""

from typing import List, Tuple, Optional
import numpy as np
from scipy.sparse import csc_matrix

from qiskit import QuantumCircuit
from qiskit_nature.operators.second_quantization import FermionicOp

from qiskit_cold_atom.base_circuit_solver import BaseCircuitSolver
from qiskit_cold_atom.exceptions import QiskitColdAtomError
from qiskit_cold_atom.fermions.fermionic_state import FermionicState
from qiskit_cold_atom.fermions.fermionic_basis import FermionicBasis
from qiskit_cold_atom.fermions.fermion_gate_library import FermionicGate


class FermionCircuitSolver(BaseCircuitSolver):
    """
    Numerically simulate fermionic systems by exactly computing the time
    evolution under unitary operations generated by fermionic Hamiltonians.
    """

    def __init__(
        self,
        shots: Optional[int] = None,
        seed: Optional[int] = None,
        num_species: int = 1,
    ):
        """
        Args:
            shots: amount of shots for the measurement simulation;
                if not None, measurements are performed
            seed: seed for the RNG for the measurement simulation
            num_species: number of different fermionic species, defaults to 1 for a single type of
                (spinless) fermions, 2 for spin-1/2 fermions etc. If > 1, the solver will check for
                conservation of the particle number per fermionic species in order to reduce the
                Hilbert space dimension of the simulation
        """
        self._basis = None
        self.num_species = num_species

        super().__init__(shots=shots, seed=seed)

    @property
    def basis(self) -> FermionicBasis:
        """
        Return the basis of fermionic occupation number states. This basis is updated via the
        setter whenever a new circuit is passed to __call__.
        """
        return self._basis

    @basis.setter
    def basis(self, basis: FermionicBasis):
        """
        Set the basis of the simulation and check its dimensions.

        Args:
            basis: The new basis.

        Raises:
            QiskitColdAtomError: If the dimension of the basis is too large.
        """
        if basis.dimension > self.max_dimension:
            raise QiskitColdAtomError(
                f"Dimension {basis.dimension} exceeds the maximum "
                f"allowed dimension {self.max_dimension}."
            )

        self._basis = basis

    def preprocess_circuit(self, circuit: QuantumCircuit):
        """
        Pre-processing fermionic circuits includes setting up the basis for the simulation
        by extracting the size, particle number and spin conservation from the circuit.

        Args:
            circuit: A fermionic quantum circuit for which to setup a basis.
        """
        initial_occupations = FermionicState.initial_state(circuit, self.num_species)
        _, spin_conservation = self._check_conservations(circuit)
        self.basis = FermionicBasis.from_state(initial_occupations, spin_conservation)
        self._dim = self.basis.dimension

    def get_initial_state(self, circuit: QuantumCircuit) -> csc_matrix:
        """
        Return the initial state of the quantum circuit as a sparse column vector.

        Args:
            circuit: The circuit for which to extract the initial_state.

        Returns:
            The initial state of the circuit as a sparse matrix.
        """

        init_state = FermionicState.initial_state(circuit, self.num_species)
        initial_occs = init_state.occupations_flat
        initial_index = self.basis.get_occupations().index(initial_occs)

        initial_state = csc_matrix(
            ([1 + 0j], ([initial_index], [0])),
            shape=(self.basis.dimension, 1),
            dtype=complex,
        )

        return initial_state

    def _embed_operator(
        self, operator: FermionicOp, num_wires: int, qargs: List[int]
    ) -> FermionicOp:
        """
        Turn a FermionicOp operator that acts on the wires given in qargs into an operator
        that acts on the entire state space of the circuit by padding with identities "I" on the
        remaining wires

        Args:
            operator: FermionicOp describing the generating Hamiltonian of a gate
            num_wires: The total number of wires in which the operator should be embedded into
            qargs: The wire indices the gate acts on

        Returns:
            FermionicOp, an operator acting on the entire quantum register of the Circuit

        Raises:
            QiskitColdAtomError:
                - If the given operator is not a FermionicOp
                - If the size of the operator does not match the given qargs
        """

        if not isinstance(operator, FermionicOp):
            raise QiskitColdAtomError(
                f"Expected FermionicOp; got {type(operator).__name__} instead."
            )

        if operator.register_length != len(qargs):
            raise QiskitColdAtomError(
                f"length of gate labels {operator.register_length} does not match "
                f"qargs {qargs} of the gates"
            )

        embedded_op_list = []

        for partial_label, factor in operator.to_list(display_format="dense"):

            full_label = ["I"] * num_wires

            for i, individual_label in enumerate(list(partial_label)):

                full_label[qargs[i]] = individual_label

            embedded_op_list.append(("".join(full_label), factor))

        return FermionicOp(embedded_op_list, display_format="dense")

    def _check_conservations(self, circuit: QuantumCircuit) -> Tuple[bool, bool]:
        """
        Check if the fermionic operators defined in the circuit conserve the total particle number
        (i.e. there are as many creation operators as annihilation operators) and the particle
        number per spin species (e.g. there are as many up/down creation operators as there are
        up/down annihilation operators).

        Args:
            circuit: A quantum circuit with fermionic gates

        Returns:
            particle_conservation: True if the particle number is conserved in the circuit
            spin_conservation: True if the particle number is conserved for each spin species

        Raises:
            QiskitColdAtomError:
                - If an operator in the circuit is not a FermionicOp.
                - If the length of the fermionic operators does not match the system size.
                - If the circuit has a number of wires that is not a multiple of the number
                  of fermionic species.
        """
        particle_conservation = True
        spin_conservation = True

        for fermionic_op in self.to_operators(circuit):

            if not isinstance(fermionic_op, FermionicOp):
                raise QiskitColdAtomError("operators need to be given as FermionicOp")

            for term in fermionic_op.to_list():
                opstring = term[0]

                if len(opstring) != circuit.num_qubits:
                    raise QiskitColdAtomError(
                        f"Expected length {circuit.num_qubits} for fermionic operator; "
                        f"received {len(opstring)}."
                    )

                num_creators = opstring.count("+")
                num_annihilators = opstring.count("-")

                if num_creators != num_annihilators:
                    return False, False

                if self.num_species > 1:

                    if circuit.num_qubits % self.num_species != 0:
                        raise QiskitColdAtomError(
                            f"The number of wires in the circuit {circuit.num_qubits} is not a "
                            f"multiple of the {self.num_species} fermionic species number."
                        )

                    sites = circuit.num_qubits // self.num_species

                    # check if the particle number is conserved for each spin species
                    for i in range(self.num_species):

                        ops = opstring[i * sites : (i + 1) * sites]
                        num_creators = ops.count("+")
                        num_annihilators = ops.count("-")

                        if num_creators != num_annihilators:
                            spin_conservation = False
                            break

        return particle_conservation, spin_conservation

    def operator_to_mat(self, operator: FermionicOp) -> csc_matrix:
        """Convert the fermionic operator to a sparse matrix.

        Args:
            operator: fermionic operator of which to compute the matrix representation

        Returns:
            scipy.sparse matrix of the Hamiltonian

        """
        return FermionicGate.operator_to_mat(operator, self.num_species, self._basis)

    def draw_shots(self, measurement_distribution: List[float]) -> List[str]:
        """
        Helper function to draw counts from a given distribution of measurement outcomes.

        Args:
            measurement_distribution: List of probabilities of the individual measurement outcomes

        Returns:
            a list of individual measurement results, e.g. ["011000", "100010", ...]
            The outcome of each shot is denoted by a binary string of the occupations of the individual
            modes in little endian convention

        Raises:
            QiskitColdAtomError:
                - If the length of the given probabilities does not match the expected Hilbert space
                dimension.
                - If the number of shots self.shots has not been specified.
        """

        meas_dim = len(measurement_distribution)

        if meas_dim != self.dim:
            raise QiskitColdAtomError(
                f"Dimension of the measurement probabilities {meas_dim} does not "
                f"match the dimension expected by the solver, {self.dim}"
            )

        if self.shots is None:
            raise QiskitColdAtomError(
                "The number of shots has to be set before drawing measurements"
            )

        # list all possible outcomes as strings '001011', reversing the order of the wires
        # to comply with Qiskit's ordering convention
        outcome_strings = ["".join(map(str, k)) for k in self.basis.get_occupations()]

        # Draw measurements:
        meas_results = np.random.choice(outcome_strings, self.shots, p=measurement_distribution)

        return meas_results.tolist()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Module to support fermionic circuits.

The fermions module holds the circuit instructions and simulators needed
by cold-atom setups that trap fermionic atoms in optical tweezer arrays or lattices.

In this setting, each wire in a quantum circuit describes a fermionic mode. Upon measurement, each mode
can be found to be occupied by a single particle (1) or be empty (0).
Backends to describe such fermionic circuits are subclasses of the :class:`BaseFermionBackend` class.
The :class:`FermionSimulator` backend is a general purpose simulator backend that simulates fermionic
circuits similar to the QasmSimulator for qubits.

Fermionic backends
-------------------
.. autosummary::
   :toctree: ../stubs/

   BaseFermionBackend
   FermionSimulator

The fermions might also come in several distinguishable species, as is the case when they carry a spin
degree of freedom. In this case, each spatial mode of an experiment can be occupied by a particle
of each spin state. In the circuit description, each individual mode is assigned its own wire.
For example, a system of spin-1/2 fermions in four spatial modes is described by a circuit with eight
wires where the first four wires denote the spin-up and the last four wires denote the spin-down modes.

Prior to applying gates, the fermionic modes in a quantum circuit need to be initialized with particles,
which defines the total number of particles (excitations) in the circuit. This initial occupation number
state can then be manipulated by applying Fermionic gates.

Fermionic gates
----------------
Fermionic gates are quantum circuit instructions designed specifically for
cold-atom based setups that control fermionic atoms in tweezers. These gates are
characterized by their effect on the fermions. Fermionic gates are subclasses or instances of the
:class:`FermionicGate` class. All of these gates define a ``generator`` property used to compute
the time-evolution. These generators are second quantized operators (:class:`FermionicOp` from Qiskit
Nature) that describe the gate Hamiltonian acting on the register.
When an entry of the :mod:`qiskit_cold_atom.fermions` module is imported, the Fermionic Gates are added
to the :class:`QuantumCircuit` class in Qiskit.

The module includes a number of gates suitable to a platform that natively implements Fermi-Hubbard type
dynamics.

.. autosummary::
   :toctree: ../stubs/

    FermionicGate
    FermiHubbard
    Hop
    Interaction
    Phase
    FRXGate
    FRYGate
    FRZGate
    LoadFermions

These gates should serve as an example of how a concrete fermionic platform can be described
through :mod:`qiskit_cold_atom`.
Users are encouraged to define their own gates to describe different fermionic hardware.
If these gates define a ``generator`` property as laid out above, the :class:`FermionSimulator`
can be used to simulate circuits with such custom gates.

Circuit Solver
----------------
Circuit solvers are classes that allow users to simulate a quantum circuit
for cold-atom based setups. They are subclasses of
:class:`BaseCircuitSolver` and can be called on quantum circuits to solve them.
The numerical simulation of fermionic circuits is carried out by the :class:`FermionCircuitSolver` class.
This simulates the circuits via exact diagonalization and provides access to the unitary,
the statevector and simulated measurement outcomes of the circuit.

.. autosummary::
   :toctree: ../stubs/

   FermionCircuitSolver

"""

from qiskit_cold_atom.fermions.fermion_simulator_backend import FermionSimulator
from qiskit_cold_atom.fermions.base_fermion_backend import BaseFermionBackend
from qiskit_cold_atom.fermions.fermion_circuit_solver import FermionCircuitSolver

from qiskit_cold_atom.fermions.fermion_gate_library import (
    FermionicGate,
    LoadFermions,
    Phase,
    Hop,
    Interaction,
    FermiHubbard,
    FRXGate,
    FRYGate,
    FRZGate,
)
